# /test/test_lint_spelling_only.py
#
# Test cases for the spellcheck-linter utility.
#
# Disable no-self-use in tests as all test methods must be
# instance methods and we don't necessarily have to use a matcher
# with them.
# pylint:disable=no-self-use
#
# See /LICENCE.md for Copyright information
"""Test the linter to ensure that each lint use-case triggers warnings."""

import atexit

import doctest

import os

import shutil

import sys

import tempfile

from iocapture import capture

from polysquarelinter import lint_spelling_only

from testtools import TestCase
from testtools.matchers import DocTestMatches


def run_with_kwargs_as_switches(func, *args, **kwargs):
    """Run :func: with :kwargs: converted to switches."""
    arguments = list(args)

    def _convert_kv_to_switches(key, value):
        """Convert a key-value pair to command-line switches."""
        append_args = ["--{0}".format(key).replace("_", "-")]

        type_dispatch = {
            bool: [],
            list: value,
            str: [value]
        }

        # We assume that the types in type_dispatch are the only types
        # we'll encounter, all others will throw an exception.
        append_args += type_dispatch[type(value)]
        return append_args

    for key, value in kwargs.items():
        arguments += _convert_kv_to_switches(key, value)

    return func(arguments)


def run_lint_spelling_only_main(filename, **kwargs):
    """Run lint_spelling_only.main with filename and kwargs as switches."""
    arguments = [filename]
    return run_with_kwargs_as_switches(lint_spelling_only.main,
                                       *arguments,
                                       **kwargs)


class TestLintSpellingOnlyAcceptance(TestCase):
    """Acceptance tests for lint_spelling_only.main()."""

    def __init__(self, *args, **kwargs):  # pylint:disable=super-on-old-class
        """Initialize class variables."""
        super(TestLintSpellingOnlyAcceptance, self).__init__(*args, **kwargs)
        self._temporary_file = None
        self._tech_words_file = None

    @classmethod
    def setUpClass(cls):  # suppress(N802)
        """Create a temporary directory to store word graph caches."""
        # This is the name of the directory that we want to
        # place our files in.
        word_cache_dir = os.path.join(os.getcwd(), "spelling_only_wordcache")
        cache_dir = tempfile.mkdtemp(prefix=word_cache_dir)
        cls.cache_dir = cache_dir
        atexit.register(shutil.rmtree, cache_dir)

    def setUp(self):  # suppress(N802)
        """Create a temporary file."""
        from six import StringIO

        super(TestLintSpellingOnlyAcceptance, self).setUp()
        self._last_directory = os.getcwd()
        self._temp_directory = tempfile.mkdtemp()
        os.chdir(self._temp_directory)
        self._temporary_file = os.path.join(os.getcwd(), "file_to_lint.txt")
        self._tech_words_file = os.path.join(os.getcwd(),
                                             "techincal_terms.txt")

        # Ensure candidate file has been created
        with open(self._temporary_file, "w"):
            pass

        with open(self._tech_words_file, "w"):
            pass

        # Patch stdout to not go anywhere
        self.patch(sys, "stdout", StringIO())

        # Finally, disable caching
        os.environ["JOBSTAMPS_DISABLED"] = "1"

    def tearDown(self):  # suppress(N802)
        """Remove temporary file."""
        os.chdir(self._last_directory)
        shutil.rmtree(self._temp_directory)
        super(TestLintSpellingOnlyAcceptance, self).tearDown()

    def _run_with_cache(self, filename, **kwargs):
        """Wrap run_lint_spelling_only_main and use caching automatically."""
        kwargs["spellcheck_cache"] = self.__class__.cache_dir
        return run_lint_spelling_only_main(filename, **kwargs)

    def test_exit_failure_on_spelling_errors(self):
        """Exit with failure on spelling errors."""
        with open(self._temporary_file, "w") as f:
            f.write("splelling error\n")

        result = self._run_with_cache(self._temporary_file)
        self.assertEqual(result, 1)

    def test_exit_failure_on_technical_word_misuse(self):
        """Exit with failure on technical word misuse."""
        with open(self._temporary_file, "w") as f:
            f.write("technical_looking_word\n")

        result = self._run_with_cache(self._temporary_file,
                                      technical_terms=self._tech_words_file)
        self.assertEqual(result, 1)

    def test_report_spelling_failure(self):
        """Report spelling failures."""
        with open(self._temporary_file, "w") as f:
            f.write("splelling error\n")

        with capture() as captured:
            self._run_with_cache(self._temporary_file)

            self.assertThat(captured.stdout,  # suppress(PYC70)
                            DocTestMatches("""... [file/spelling_error] ...""",
                                           doctest.ELLIPSIS |
                                           doctest.NORMALIZE_WHITESPACE |
                                           doctest.REPORT_NDIFF))

    def test_report_technical_word_misuse(self):
        """Report technical word misuse."""
        with open(self._temporary_file, "w") as f:
            f.write("technical_looking_word\n")

        with capture() as captured:
            self._run_with_cache(self._temporary_file,
                                 technical_terms=self._tech_words_file)

            self.assertThat(captured.stdout,  # suppress(PYC70)
                            # suppress(file/spelling_error)
                            DocTestMatches("""... technical_looking_word"""
                                           """...""",
                                           doctest.ELLIPSIS |
                                           doctest.NORMALIZE_WHITESPACE |
                                           doctest.REPORT_NDIFF))

    def test_exit_success_on_technical_word_found_in_list(self):
        """Exit with success when a technical word is found."""
        with open(self._temporary_file, "w") as f:
            f.write("technical_looking_word\n")

        with open(self._tech_words_file, "w") as f:
            f.write("technical_looking_word\n")

        result = self._run_with_cache(self._temporary_file,
                                      technical_terms=self._tech_words_file)
        self.assertEqual(result, 0)

    def test_allow_technical_words_in_user_dictionary(self):
        """Allow technical terms in user dictionary."""
        with open(os.path.join(os.getcwd(), "DICTIONARY"), "w") as f:
            f.write("technical_looking_word\n")

        with open(self._tech_words_file, "w") as f:
            f.write("technical_looking_word\n")

        result = self._run_with_cache(self._temporary_file)

        self.assertEqual(result, 0)



# /polysquarelinter/spelling.py
#
# Implements spell-checking functionality for polysquarelinter.
#
# See /LICENCE.md for Copyright information
"""Implements spell-checking functionality for polysquarelinter.

The main two functions of interest are
spellcheckable_and_shadow_contents and spellcheck_region.

spellcheckable_and_shadow_contents will return a tuple
of both a list of _ChunkInfo classes and a shadow region.

The _ChunkInfo classes, represent a region spanning some lines
of text. The data member of _ChunkInfo is a list of lines covering
that region.

The shadow region is an array of characters with 1-1 mapping on the rest
of the input lines which were not covered by any chunk. This can
be used, for example, to generate a list of "technical words" used
in the rest of the file. technical_words_from_shadow_contents provides
this functionality as a convenience.

spellcheck_region takes a list of lines and spell-checks every
word in that region. If it detects an error it will return
a SpellcheckError indicating the offset into that chunk
the error was detected.
"""

import errno

import functools

import itertools

import os

import re

from collections import namedtuple

from pkg_resources import resource_stream

from whoosh import spelling
from whoosh.automata import fst
from whoosh.filedb.filestore import FileStorage, RamStorage, copy_to_ram

_SPELLCHECKABLE_WORDS = r"^([A-Za-z][a-z']*|[A-Z']*)$"
_VALID_SYMBOL_WORDS = r"^[A-Za-z_][A-Za-z0-9_\.]*$"

# These are intended to be used as "global" variables, as caches. We
# don't want to expose these caches to the user. As such, they should
# not be named with the constant naming convention.
_spellchecker_cache = dict()  # suppress(invalid-name)
_valid_words_cache = dict()  # suppress(invalid-name)
_user_dictionary_cache = dict()  # suppress(invalid-name)


def clear_caches():  # suppress(unused-function)
    """Clear all caches."""
    for _, reader in _spellchecker_cache.values():
        reader.close()

    _spellchecker_cache.clear()
    _valid_words_cache.clear()
    _user_dictionary_cache.clear()


FileCommentSystem = namedtuple("FileCommentSystem", "begin middle end")


def _comment_system_for_file(contents):
    """For file contents, return the comment system."""
    if contents[0] == "#":
        return FileCommentSystem(begin="#", middle="", end="")
    elif contents[:2] == "/*":
        return FileCommentSystem(begin="/*", middle="*", end="*/")
    elif contents[:2] == "//":
        return FileCommentSystem(begin="//", middle="//", end="")
    elif contents[:3] == "rem":
        return FileCommentSystem(begin="rem", middle="rem", end="")
    else:
        raise RuntimeError("Couldn't detect comment "
                           "system from {0}".format(contents[:3]))


@functools.total_ordering
class _Marker(namedtuple("Marker", "line col")):
    """A marker for a point in a text file. Comparable."""

    def __lt__(self, other):
        """True if self is less than other."""
        if self.line == other.line:
            return self.col < other.col

        return self.line < other.line


@functools.total_ordering
class _ChunkInfo(namedtuple("_ChunkInfo", "line column data type")):
    """A chunk of selected text, starting at line/column."""

    # The chunk type.
    #
    # A Shadow chunk exists only to filter out other chunks by virtue
    # of being "first". It shouldn't be considered a region of selected
    # text.
    #
    # A real chunk is a region of selected text.
    Real = 0
    Shadow = 1

    def start(self):
        """Start point of this chunk, as a Marker."""
        return _Marker(line=self.line, col=self.column)

    def end(self):
        """End point of this chunk, as a Marker."""
        if len(self.data) == 1:
            end_col = self.column + len(self.data[0])
        else:
            end_col = len(self.data[-1])

        return _Marker(line=self.line + len(self.data) - 1, col=end_col)

    def __lt__(self, other):
        """True if self is less than other."""
        if self.line == other.line:
            if self.column == other.column:
                # Shadow chunks always end up after real chunks. This
                # will cause them to be excluded under the
                # first-chunk-wins rule.
                return self.type < other.type

            return self.column < other.column

        return self.line < other.line


def _get_in_out_points(lines, chunk_begin_regex, chunk_end_regex):
    """Get a set of line/column markers for each in and out point.

    The in points match :chunk_begin_regex: and the out points match
    :chunk_end_regex:. Note that this function might return duplicated points
    or in and out points which are otherwise invalid. They should be
    filtered such that there are no duplicated points and no repeated
    in points before out points.
    """
    all_in_points = []
    all_out_points = []

    for index, line in enumerate(lines):
        line_out_points = []
        line_in_points = [m.end() for m in re.finditer(chunk_begin_regex,
                                                       line)]

        if not chunk_end_regex:
            if len(line_in_points):
                line_out_points = [len(line)]
        else:
            line_out_points = [m.start() for m in re.finditer(chunk_end_regex,
                                                              line)]

        all_in_points.extend([_Marker(index, c) for c in line_in_points])
        all_out_points.extend([_Marker(index, c) for c in line_out_points])

    return all_in_points, all_out_points


def _find_chunks(lines, chunk_begin_regex, chunk_end_regex, chunk_type):
    """Find chunks delimited by chunk_begin_regex and chunk_end_regex.

    The found chunks are returned as part of a _ChunkInfo struct which
    contains details about the line and character offsets where the chunk
    starts and finishes.

    Note that this function may return overlapping chunks. To avoid that,
    filter the chunks through _filter_overlapping_chunks.

    The "data" member is a list of strings, each going line-by-line.

    The chunk will be assigned a type of :chunk_type:
    """
    all_in_points, all_out_points = _get_in_out_points(lines,
                                                       chunk_begin_regex,
                                                       chunk_end_regex)

    # Filter out alternating in and out points if they are denoted
    # by the same marker.
    if chunk_end_regex == chunk_begin_regex:
        assert len(all_out_points) == len(all_in_points)
        length = len(all_out_points)
        all_out_points = [all_out_points[i] for i in range(1, length, 2)]
        all_in_points = [all_in_points[i] for i in range(0, length, 2)]

    # Now use all the in and out points to generate _ChunkInfo structs
    while len(all_out_points):
        # Move to next in/out points
        in_point = all_in_points.pop(0)
        out_point = all_out_points.pop(0)

        # Move in and out points along such that
        # !((in_point | out_point) & shadow_contents) and
        # in_point > out_point
        while len(all_in_points) and all_in_points[0] <= out_point:
            all_in_points.pop(0)

        while len(all_out_points) and out_point < in_point:
            out_point = all_out_points.pop(0)

        # Nothing here, move on
        if in_point == out_point:
            continue

        assert out_point > in_point

        chunk_data = []
        for line_index in range(in_point.line, out_point.line + 1):
            line = len(lines[line_index])
            in_col = in_point.col if line_index == in_point.line else 0
            out_col = out_point.col if line_index == out_point.line else line

            chunk_data.append(lines[line_index][in_col:out_col])

        yield _ChunkInfo(in_point.line,
                         in_point.col,
                         chunk_data,
                         chunk_type)


def _filter_overlapping_chunks(chunks):
    """Return a list of non-overlapping chunks on a first chunk wins basis.

    A chunk is considered to be overlapping if it starts before a chunk
    prior to it ends. We keep the former chunk.
    """
    chunks = sorted(chunks)
    if len(chunks) < 2:
        return chunks

    # Always take the first chunks
    filtered_chunks = [chunks[0]]

    compare_index = 0
    take_index = 1
    while take_index < len(chunks):
        # Advance take_index if the chunk it refers to starts
        # before the compare_index chunk end point. If this is
        # false, then the chunks do not overlap. As such, assign
        # compare_index to take_index and continue.
        #
        # This is highly sensitive to the sorting order of the chunks.
        # Because docstrings also count as a chunk and will count
        # in the same place as a regular "shadow" string, the shadow
        # string must appear after the docstring does. This will
        # cause the shadow string to be eliminated.
        if chunks[take_index].start() > chunks[compare_index].end():
            filtered_chunks.append(chunks[take_index])
            compare_index = take_index

        take_index += 1

    return filtered_chunks


def _split_line_with_offsets(line):
    """Split a line by delimiter, but yield tuples of word and offset.

    This function works by dropping all the english-like punctuation from
    a line (so parenthesis preceded or succeeded by spaces, periods, etc)
    and then splitting on spaces.
    """
    for delimiter in re.finditer(r"[\.,:\;](?![^\s])", line):
        span = delimiter.span()
        line = line[:span[0]] + " " + line[span[1]:]

    for delimiter in re.finditer(r"[\"'\)\]\}>](?![^\.,\;:\"'\)\]\}>\s])",
                                 line):
        span = delimiter.span()
        line = line[:span[0]] + " " + line[span[1]:]

    for delimiter in re.finditer(r"(?<![^\.,\;:\"'\(\[\{<\s])[\"'\(\[\{<]",
                                 line):
        span = delimiter.span()
        line = line[:span[0]] + " " + line[span[1]:]

    # Treat hyphen separated words as separate words
    line = line.replace("-", " ")

    # Remove backticks
    line = line.replace("`", " ")

    for match in re.finditer(r"[^\s]+", line):
        content = match.group(0)
        if content.strip() != "":
            yield (match.span()[0], content)


def read_dictionary_file(dictionary_path):
    """Return all words in dictionary file as set."""
    try:
        return _user_dictionary_cache[dictionary_path]
    except KeyError:
        if dictionary_path and os.path.exists(dictionary_path):
            with open(dictionary_path, "rt") as dict_f:
                words = set(re.findall(r"(\w[\w']*\w|\w)",
                                       " ".join(dict_f.read().splitlines())))
                return words

        return set()


def valid_words_set(path_to_user_dictionary=None,
                    user_dictionary_words=None):
    """Get a set of valid words.

    If :path_to_user_dictionary: is specified, then the newline-separated
    words in that file will be added to the word set.
    """
    def read_file(binary_file):
        """Read a binary file for its text lines."""
        return binary_file.read().decode("ascii").splitlines()

    try:
        valid = _valid_words_cache[path_to_user_dictionary]
        return valid
    except KeyError:
        words = set()
        with resource_stream("polysquarelinter", "en_US.txt") as words_file:
            words |= set(["".join(l).lower() for l in read_file(words_file)])

        if path_to_user_dictionary:
            # Add both case-sensitive and case-insensitive variants
            # of words in user dictionary as they may be checked as
            # though they are a regular word and a technical word.
            words |= set([w.lower() for w in user_dictionary_words])
            words |= user_dictionary_words

        _valid_words_cache[path_to_user_dictionary] = words
        return words


# Store the corrector and reader in the cache
SpellcheckerCacheEntry = namedtuple("SpellcheckerCacheEntry",
                                    "corrector reader")


def _create_word_graph_file(name, file_storage, word_set):
    """Create a word graph file and open it in memory."""
    word_graph_file = file_storage.create_file(name)
    spelling.wordlist_to_graph_file(sorted(list(word_set)),
                                    word_graph_file)
    return copy_to_ram(file_storage).open_file(name)


def _spellchecker_for(word_set,
                      name,
                      spellcheck_cache_path=None,
                      sources=None):
    """Get a whoosh spellchecker for :word_set:.

    The word graph for this spellchecker will be stored on-disk with
    the unique-name :name: in :spellcheck_cache_path:, if it exists.
    This allows for much faster loading of word graphs after they have been
    pre-populated.

    :sources: is a list of filenames which will be checked to see if they
    are newer than the stored word graph. If they are newer, then the word
    graph gets repopulated.
    """
    if _spellchecker_cache.get(name, None) is not None:
        return _spellchecker_cache[name].corrector

    # Check the modification time of all the paths in :sources: to see
    # if they've been modified since the cache file was created. If so,
    # delete the cache file. This will cause it to be regenerated.
    #
    # Note that this relies on an implementation detail in whoosh, namely
    # that the cache file is always stored at spellcheck_cache_path/name.
    if spellcheck_cache_path:

        # Ensure that the directory has been created
        try:
            os.makedirs(spellcheck_cache_path)
        except OSError as error:
            if error.errno != errno.EEXIST:  # suppress(PYC90)
                raise error

        graph_path = os.path.realpath(spellcheck_cache_path)
        file_storage = FileStorage(graph_path)

        preexisting_cache = os.path.abspath(os.path.join(spellcheck_cache_path,
                                                         name))
        if os.path.exists(preexisting_cache):
            cache_mtime = os.path.getmtime(preexisting_cache)
            for source in sources:
                source_path = os.path.realpath(source)
                if not os.path.exists(source_path):
                    continue

                if os.path.getmtime(source_path) > cache_mtime:
                    file_storage.delete_file(name)
                    break

        try:
            word_graph = copy_to_ram(file_storage).open_file(name)
        except (IOError, NameError):
            word_graph = _create_word_graph_file(name, file_storage, word_set)

    else:
        ram_storage = RamStorage()
        word_graph = _create_word_graph_file(name, ram_storage, word_set)

    reader = fst.GraphReader(word_graph)
    corrector = spelling.GraphCorrector(reader)
    _spellchecker_cache[name] = SpellcheckerCacheEntry(corrector, reader)
    return corrector


def filter_nonspellcheckable_tokens(line, block_out_regexes=None):
    """Return line with paths, urls and emails filtered out.

    Block out other strings of text matching :block_out_regexes: if passed in.
    """
    all_block_out_regexes = [
        r"[^\s]*:[^\s]*[/\\][^\s]*",
        r"[^\s]*[/\\][^\s]*",
        r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Za-z]+\b"
    ] + (block_out_regexes or list())

    for block_regex in all_block_out_regexes:
        for marker in re.finditer(block_regex, line):
            spaces = " " * (marker.end() - marker.start())
            line = line[:marker.start()] + spaces + line[marker.end():]

    return line


def _shadow_contents_from_chunks(contents, chunks, block_out_regexes=None):
    """Remove all contents in spellcheckable :chunks: from contents."""
    shadow_contents = [list(l) for l in contents]
    for chunk in chunks:
        char_offset = chunk.column
        line_offset = 0

        for index, line in enumerate(chunk.data):
            # Block out entire chunk range from shadow_contents
            for character_in_line in range(0, len(line)):
                shadow_line = chunk.line + line_offset
                shadow_char = char_offset + character_in_line
                shadow_contents[shadow_line][shadow_char] = 0

            # Also block out certain regexps from this chunk
            line = filter_nonspellcheckable_tokens(line,
                                                   block_out_regexes)
            chunk.data[index] = line

            line_offset += 1
            char_offset = 0

    return shadow_contents


# There doesn't seem to be a shorter name that can be given to this function
# which also retains its descriptive value. The function return both
# the spellcheckable contents and the remaining contents which are not
# spellcheckable.
#
# suppress(invalid-name)
def spellcheckable_and_shadow_contents(contents, block_out_regexes=None):
    """For contents, split into spellcheckable and shadow parts.

    :contents: is a list of lines in a file.

    The return value is a tuple of (chunks, shadow_contents).
    chunks is a list of _ChunkInfo, each of which contain
    a region of text to be spell-checked. shadow_contents is an array of
    characters and integers. The characters represent nonspellcheckable
    regions and any region which will be subject to spellcheck is denoted
    by a zero in place of that character.
    """
    comment_system = _comment_system_for_file(contents[0])
    begin_comment = comment_system.begin.replace("*", r"\*")

    if comment_system.end == "":
        end_comment = None
    else:
        end_comment = comment_system.end.replace("*", r"\*")

    iterables = [
        _find_chunks(contents,
                     r"(?<![^\\]\\)\"",
                     r"(?<![^\\]\\)\"",
                     _ChunkInfo.Shadow),
        _find_chunks(contents,
                     r"(?<![^\\]\\)'",
                     r"(?<![^\\]\\)'",
                     _ChunkInfo.Shadow),
        _find_chunks(contents, r"\"\"\"", r"\"\"\"", _ChunkInfo.Real),
        _find_chunks(contents, r"'''", r"'''", _ChunkInfo.Real),
        _find_chunks(contents, begin_comment, end_comment, _ChunkInfo.Real)
    ]

    # Shadow contents excludes anything in quotes
    chunks = _filter_overlapping_chunks(list(itertools.chain(*iterables)))
    shadow_contents = _shadow_contents_from_chunks(contents,
                                                   chunks,
                                                   block_out_regexes)

    # Only return chunks which are actually spellcheckable and not just
    # quotes which are intended to block out parts of the shadow contents
    chunks = [c for c in chunks if c.type == _ChunkInfo.Real]
    return (chunks, shadow_contents)


def _split_into_symbol_words(sym):
    """Split a technical looking word into a set of symbols.

    This handles cases where technical words are separated by dots or
    arrows, as is the convention in many programming languages.
    """
    punc = r"[\s\-\*/\+\.,:\;=\)\(\[\]\{\}<>\|\?&\^\$@]"
    words = [w.strip() for w in re.split(punc, sym)]
    return words


# There is no shorter function name which will cause this function
# name to lose its descriptive value.
#
# suppress(invalid-name)
def technical_words_from_shadow_contents(shadow_contents):
    """Get a set of technical words from :shadow_contents:.

    :shadow_contents: is an array of shadow contents, as returned by
    spellcheckable_and_shadow_contents.
    """
    technical_words = set()
    for line in shadow_contents:
        # "Fix up" the shadow line by replacing zeros with spaces.
        line = "".join([(lambda c: " " if c == 0 else c)(c) for c in line])
        for sym in _split_into_symbol_words(line):
            if len(sym) and re.compile(_VALID_SYMBOL_WORDS).match(sym):
                technical_words |= set([sym])

    return technical_words


@functools.total_ordering
class SpellcheckError(namedtuple("SpellcheckError",
                                 "word "
                                 "line_offset "
                                 "column_offset "
                                 "suggestions "
                                 "error_type")):
    """A spelling error, relative to a certain region of lines."""

    InvalidWord = 0
    TechnicalWord = 1

    def __lt__(self, other):
        """True if self is less than other."""
        if self.line_offset == other.line_offset:
            return self.column_offset < other.column_offset

        return self.line_offset < other.line_offset


def _error_if_word_invalid(word,
                           valid_words_dictionary,
                           technical_words_dictionary,
                           line_offset,
                           col_offset):
    """Return SpellcheckError if this non-technical word is invalid."""
    word_lower = word.lower()
    valid_words_result = valid_words_dictionary.corrections(word_lower)

    if technical_words_dictionary:
        technical_words_result = technical_words_dictionary.corrections(word)
    else:
        # No technical words available to make an otherwise invalid
        # result value.
        technical_words_result = Dictionary.Result(False, list())

    if not valid_words_result.valid and not technical_words_result.valid:
        return SpellcheckError(word,
                               line_offset,
                               col_offset,
                               valid_words_result.suggestions,
                               SpellcheckError.InvalidWord)


def _error_if_symbol_unused(symbol_word,
                            technical_words_dictionary,
                            line_offset,
                            col_offset):
    """Return SpellcheckError if this symbol is not used in the code."""
    result = technical_words_dictionary.corrections(symbol_word,
                                                    distance=5,
                                                    prefix=0)
    if not result.valid:
        return SpellcheckError(symbol_word,
                               line_offset,
                               col_offset,
                               result.suggestions,
                               SpellcheckError.TechnicalWord)


class Dictionary(object):
    """A dictionary which can find corrections a word set.

    Customize the words available in the dictionary by passing a set of
    words as :words: .

    :dictionary_sources: is list of paths to text files where :words: set
    was sourced from. If :cache: is set, then the dictionary will be loaded
    from an on-disk binary cache loaded inside the path specified by
    :cache: , but will be regenerated if the cache file is older than
    than any of the paths in :dictionary_sources:

    """

    def __init__(self, words, name, dictionary_sources=None, cache=None):
        """Initialize this Dictionary.

        This sets the dictionary's words and creates a corrector
        object, from a cache if possible.
        """
        super(Dictionary, self).__init__()
        self._words = words
        self._corrector = _spellchecker_for(words,
                                            name,
                                            cache,
                                            sources=dictionary_sources)

    Result = namedtuple("Result", "valid suggestions")

    def corrections(self, word, prefix=1, distance=2):
        """Get corrections for word, if word is an invalid word.

        :prefix: is the number of characters the prefix of the word must
        have in common with the suggested corrections.

        :distance: is the character distance the corrections may have between
        the input word. This limits the number of available corrections
        but decreases the correction search space.

        The return value of this function is a Result tuple, with the
        :valid: member indicating whether the input word is a valid one and
        :suggestions: member containing a list of suggestions.
        """
        if word not in self._words:
            return Dictionary.Result(False,
                                     self._corrector.suggest(word,
                                                             prefix=prefix,
                                                             maxdist=distance))
        else:
            return Dictionary.Result(True, list())

    def words(self):
        """Retrieve set of words in this dictionary."""
        return self._words


def spellcheck_region(region_lines,
                      valid_words_dictionary=None,
                      technical_words_dictionary=None,
                      user_dictionary_words=None):
    """Perform spellcheck on each word in :region_lines:.

    Each word will be checked for existence in :valid_words_dictionary:.
    If it is not in :valid_words_dictionary: then corrections will be
    suggested.

    If the word isn't one which is an ordinary word, then it will be checked
    against the available symbols in :technical_words_dictionary: . If it is
    not in :technical_words_dictionary: then corrections will be suggested.
    """
    user_dictionary_words = user_dictionary_words or set()
    spellcheckable_words_regex = re.compile(_SPELLCHECKABLE_WORDS)

    line_offset = 0
    for line in region_lines:
        for col_offset, word in _split_line_with_offsets(line):
            word = word.strip()
            if len(word) == 0:
                continue

            # If this word exists in the user dictionary, then always allow
            # it, even if it might be technical in nature
            if word in user_dictionary_words:
                continue

            if (valid_words_dictionary and
                    spellcheckable_words_regex.match(word)):
                error = _error_if_word_invalid(word,
                                               valid_words_dictionary,
                                               technical_words_dictionary,
                                               line_offset,
                                               col_offset)

                if error:
                    yield error

            # Check for symbols appearing in comments or
            # docstrings.
            elif technical_words_dictionary:
                for symbol_word in _split_into_symbol_words(word):
                    if not re.compile(_VALID_SYMBOL_WORDS).match(symbol_word):
                        continue

                    error = _error_if_symbol_unused(symbol_word,
                                                    technical_words_dictionary,
                                                    line_offset,
                                                    col_offset)

                    if error:
                        yield error

        line_offset += 1



# /polysquarelinter/linter.py
#
# Entry point for linter.
#
# See /LICENCE.md for Copyright information
"""Main module for linter."""

import argparse

import hashlib

import itertools

import multiprocessing

import os

import re

import sys

import tempfile

import traceback

from collections import OrderedDict, defaultdict, namedtuple

from contextlib import closing

from functools import reduce as freduce

from jobstamps import jobstamp

import parmap

from polysquarelinter.spelling import (Dictionary,
                                       SpellcheckError,
                                       read_dictionary_file,
                                       spellcheck_region,
                                       spellcheckable_and_shadow_contents,
                                       technical_words_from_shadow_contents,
                                       valid_words_set)

try:
    from Queue import Queue
except ImportError:
    # suppress(F811,E301,E101,F401,unused-import,import-error)
    from queue import Queue

_ALL_COMMENT = r"(/\*| \*\/| \*|#|//|rem)"
_HDR_COMMENT = r"(/\*|#|//|rem)"
_MDL_COMMENT = r"( \*\/| \*|#|//|rem)"
_FTR_COMMENT = r"(/\*| \*)"


def _comment_type_from_line(line):
    """Return the "comment header" (' * ', '# ', 'rem ', '// ', '/* ').

    This header goes before the content of a start of the
    line in a replacement.
    """
    regex = re.compile(r"^{0}".format(_ALL_COMMENT))
    match = regex.match(line)
    if match:
        return "{0} ".format(line[match.start():match.end()])

    raise RuntimeError("Unable to find comment header for {0}".format(line))


def _end_comment_type_from_line(line):
    """Return the "comment footer" (eg ' */'').

    This header goes after the content of a start of the
    line in a replacement that would be at the end of a header block.
    """
    regex = re.compile(r"^{0}".format(_FTR_COMMENT))
    match = regex.match(line)
    if match:
        return " */"

    return ""


LinterFailure = namedtuple("LinterFailure", "description line replacement")


def _line_is_shebang(line):
    """Return true if line is a shebang."""
    regex = re.compile(r"^(#!|@echo off).*$")
    if regex.match(line):
        return True

    return False


def _filename_in_headerblock(relative_path, contents, linter_options):
    """Check for a filename in a header block.

    like such:
    # /path/to/filename
    """
    del linter_options
    check_index = 0

    if len(contents) > 0:
        if _line_is_shebang(contents[0]):
            check_index = 1

    if len(contents) < check_index + 1:
        description = ("""Document cannot have less than """
                       """{0} lines""").format(check_index + 1)
        return LinterFailure(description, 1, replacement=None)

    header_path = relative_path.replace("\\", "/")
    regex = re.compile(r"^{0} \/{1}$".format(_HDR_COMMENT,
                                             re.escape(header_path)))
    if not regex.match(contents[check_index]):
        description = ("""The filename /{0} must be the """
                       """first line of the header""")
        return LinterFailure(description.format(header_path),
                             check_index + 1,
                             _comment_type_from_line(contents[check_index]) +
                             "/{0}\n".format(header_path))


def _match_space_at_line(line):
    """Return a re.match object if an empty comment was found on line."""
    regex = re.compile(r"^{0}$".format(_MDL_COMMENT))
    return regex.match(line)


def _space_in_headerblock(relative_path, contents, linter_options):
    """Check for space between the filename in a header block and description.

    like such:
    # /path/to/filename
    #
    # Description
    """
    del relative_path
    del linter_options

    check_index = 1

    if len(contents) > 0:
        if _line_is_shebang(contents[0]):
            check_index = 2

    if len(contents) < check_index + 1:
        description = ("""Document cannot have less """
                       """than {0} lines""").format(check_index + 1)
        return LinterFailure(description, 1, replacement=None)

    candidate = contents[check_index]

    if not _match_space_at_line(candidate):
        description = """The second line must be an empty comment"""
        return LinterFailure(description, check_index + 1,
                             _comment_type_from_line(candidate)[:-1] + "\n" +
                             candidate)


def _find_last_line_index(contents):
    """Find the last line of the headerblock in contents."""
    lineno = 0
    headerblock = re.compile(r"^{0}.*$".format(_ALL_COMMENT))
    while headerblock.match(contents[lineno]):
        if lineno + 1 == len(contents):
            raise RuntimeError("""No end of headerblock in file""")
        lineno = lineno + 1

    if lineno < 2:
        raise RuntimeError("""Headerblock must have at least two lines""")

    return lineno - 1


def _space_before_copyright(relative_path, contents, linter_options):
    """Check for a space between the last line and description.

    like such
    # Description
    #
    # Last Line
    """
    del relative_path
    del linter_options

    last_line = _find_last_line_index(contents)
    if not _match_space_at_line(contents[last_line - 1]):
        comment = _comment_type_from_line(contents[last_line])[:-1]
        description = """The second last line must be an empty comment"""
        return LinterFailure(description, last_line + 1,
                             comment + "\n" + contents[last_line])


def _copyright_end_of_headerblock(relative_path, contents, linter_options):
    """Check for copyright notice at end of headerblock."""
    del relative_path
    del linter_options

    lineno = _find_last_line_index(contents)
    notice = "See /LICENCE.md for Copyright information"
    regex = re.compile(r"^{0} {1}( .*$|$)".format(_MDL_COMMENT, notice))
    if not regex.match(contents[lineno]):
        description = ("""The last of the header block line must have the """
                       """following notice: {0}""")
        replacement = None
        comment = _comment_type_from_line(contents[lineno])
        comment_footer = _end_comment_type_from_line(contents[lineno])
        # If the last line has the words "Copyright" or "/LICENCE.md" the
        # user probably attempted to add a notice, but failed, so just
        # suggest replacing the whole line
        if re.compile(r"^.*(Copyright|LICENCE.md).*$").match(contents[lineno]):
            replacement = "{0}{1}\n".format(comment, notice + comment_footer)
        # Put the copyright notice on a new line
        else:
            # Strip off the footer and the \n at the end of the line.
            repl_contents = contents[lineno][:-(len(comment_footer) + 1)]
            replacement = "{0}\n{1}{2}\n".format(repl_contents,
                                                 comment,
                                                 notice + comment_footer)

        return LinterFailure(description.format(notice),
                             lineno + 1,
                             replacement)


def _newline_end_of_file(relative_path, contents, linter_options):
    r"""Check that every file ends with a single \n."""
    del relative_path
    del linter_options
    last_line = len(contents) - 1
    if not contents[last_line].endswith("\n"):
        return LinterFailure(r"No \n at end of file",
                             last_line + 1,
                             "{0}\n".format(contents[last_line]))


def _no_trailing_whitespace(relative_path, contents, linter_options):
    r"""Check that every file ends with a single \n."""
    del relative_path
    del linter_options

    trailing_whitespace = re.compile(r"(?<![ \t\r\n])[ \t]+$")

    for index, line in enumerate(contents):
        for space in trailing_whitespace.finditer(line):
            assert space.end() == len(line) - 1
            return LinterFailure("Trailing whitespace",
                                 index + 1,
                                 line[:space.start()] + line[space.end():])


# Disabling too-many-arguments as this is just a helper function
# to populate another named tuple.
#
# suppress(too-many-arguments)
def _populate_spelling_error(word,
                             suggestions,
                             contents,
                             line_offset,
                             column_offset,
                             message_start):
    """Create a LinterFailure for word.

    This function takes suggestions from :suggestions: and uses it to
    populate the message and candidate replacement. The replacement will
    be a line in :contents:, as determined by :line_offset: and
    :column_offset:.
    """
    error_line = contents[line_offset]
    if len(suggestions):
        char_word_offset = (column_offset + len(word))
        replacement = (error_line[:column_offset] +
                       suggestions[0] +
                       error_line[char_word_offset:])
    else:
        replacement = None

    if len(suggestions):
        suggestions_text = (""", perhaps you meant """ +
                            " ".join(suggestions))
    else:
        suggestions_text = ""

    format_desc = message_start + suggestions_text
    return LinterFailure(format_desc,
                         line_offset + 1,
                         replacement)


def _determine_character_offset(line_offset, character_offset, chunk_offset):
    """Apply chunk_offset if line_offset is zero."""
    if line_offset == 0:
        return character_offset + chunk_offset
    else:
        return character_offset


_SPELLCHECK_MESSAGES = {
    SpellcheckError.InvalidWord: """Misspelled word {0}""",
    SpellcheckError.TechnicalWord: """Technical word {0} not used in """
                                   """surrounding body """
}


def _find_spelling_errors_in_chunks(chunks,
                                    contents,
                                    valid_words_dictionary=None,
                                    technical_words_dictionary=None,
                                    user_dictionary_words=None):
    """For each chunk and a set of valid and technical words, find errors."""
    for chunk in chunks:
        for error in spellcheck_region(chunk.data,
                                       valid_words_dictionary,
                                       technical_words_dictionary,
                                       user_dictionary_words):
            col_offset = _determine_character_offset(error.line_offset,
                                                     error.column_offset,
                                                     chunk.column)
            msg = _SPELLCHECK_MESSAGES[error.error_type].format(error.word)
            yield _populate_spelling_error(error.word,
                                           error.suggestions,
                                           contents,
                                           error.line_offset +
                                           chunk.line,
                                           col_offset,
                                           msg)


def _create_valid_words_dictionary(spellchecker_cache_path):
    """Create a Dictionary at spellchecker_cache_path with valid words."""
    user_dictionary = os.path.join(os.getcwd(), "DICTIONARY")
    user_words = read_dictionary_file(user_dictionary)

    valid_words = Dictionary(valid_words_set(user_dictionary, user_words),
                             "valid_words",
                             [user_dictionary],
                             spellchecker_cache_path)

    return (user_words, valid_words)


# suppress(invalid-name)
def _create_technical_words_dictionary(spellchecker_cache_path,
                                       relative_path,
                                       user_words,
                                       shadow):
    """Create Dictionary at spellchecker_cache_path with technical words."""
    technical_terms_set = (user_words |
                           technical_words_from_shadow_contents(shadow))
    technical_words = Dictionary(technical_terms_set,
                                 "technical_words_" +
                                 relative_path.replace(os.path.sep, "_"),
                                 [os.path.realpath(relative_path)],
                                 spellchecker_cache_path)
    return technical_words


def _construct_user_dictionary(global_options, tool_options):
    """Cause dictionary with valid and user words to be cached on disk."""
    del global_options

    spellchecker_cache_path = tool_options.get("spellcheck_cache", None)
    _create_valid_words_dictionary(spellchecker_cache_path)


def _drain(queue_to_drain, sentinel=None):
    """Remove all values from queue_to_drain and return as list.

    This uses the trick from
    http://stackoverflow.com/questions/
    1540822/dumping-a-multiprocessing-queue-into-a-list in order to ensure that
    the queue is fully drained.
    """
    queue_to_drain.put(sentinel)
    queued_items = [i for i in iter(queue_to_drain.get, None)]
    return queued_items


def _maybe_log_technical_terms(global_options, tool_options):
    """Log technical terms as appropriate if the user requested it.

    As a side effect, if --log-technical-terms-to is passed to the linter
    then open up the file specified (or create it) and then merge the set
    of technical words that we have now with the technical words already
    in it.
    """
    log_technical_terms_to_path = global_options.get("log_technical_terms_to",
                                                     None)
    log_technical_terms_to_queue = tool_options.get("log_technical_terms_to",
                                                    None)
    if log_technical_terms_to_path:
        assert log_technical_terms_to_queue is not None

        if not log_technical_terms_to_queue.empty():
            with closing(os.fdopen(os.open(log_technical_terms_to_path,
                                           os.O_RDWR | os.O_CREAT),
                                   "r+")) as terms_file:
                # pychecker can't see through the handle returned by closing
                # so we need to suppress these warnings.
                terms = set(terms_file.read().splitlines())
                new_terms = set(freduce(lambda x, y: x | y,
                                        _drain(log_technical_terms_to_queue)))

                if not terms.issuperset(new_terms):
                    terms_file.seek(0)
                    terms_file.truncate(0)
                    terms_file.write("\n".join(list(terms |
                                                    set(new_terms))))


def _no_spelling_errors(relative_path, contents, linter_options):
    """No spelling errors in strings, comments or anything of the like."""
    block_regexps = linter_options.get("block_regexps", None)
    chunks, shadow = spellcheckable_and_shadow_contents(contents,
                                                        block_regexps)
    cache = linter_options.get("spellcheck_cache", None)
    user_words, valid_words = _create_valid_words_dictionary(cache)
    technical_words = _create_technical_words_dictionary(cache,
                                                         relative_path,
                                                         user_words,
                                                         shadow)

    if linter_options.get("log_technical_terms_to"):
        linter_options["log_technical_terms_to"].put(technical_words.words())

    return [e for e in _find_spelling_errors_in_chunks(chunks,
                                                       contents,
                                                       valid_words,
                                                       technical_words,
                                                       user_words) if e]


def _line_suppresses_error_code(line, code):
    """Check if line contains necessary content to suppress code.

    A line suppresses code if it is in the format suppress(code1,code2) etc.
    """
    match = re.compile(r"suppress\((.*)\)").match(line)
    if match:
        codes = match.group(1).split(",")
        return code in codes

    return False


def _error_is_suppressed(error, code, contents):
    """Return true if error is suppressed by an inline suppression."""
    if len(contents) == 0:
        return False

    if error.line > 1:
        # Check above, and then to the side for suppressions
        above = contents[error.line - 2].split("#")
        if len(above) and _line_suppresses_error_code(above[-1].strip(),
                                                      code):
            return True

    aside = contents[error.line - 1].split("#")
    if len(aside) and _line_suppresses_error_code(aside[-1].strip(), code):
        return True

    return False


class _LinterFunction(namedtuple("_LinterFunction",
                                 "function before_all after_all")):
    """A function and some before/after hooks to be called."""

    def __repr__(self):
        """Print function names, but no more."""
        func_name = self.function.__name__ if self.function else "_"
        before_name = self.before_all.__name__ if self.before_all else "_"
        after_name = self.after_all.__name__ if self.after_all else "_"

        return "_LinterFunction({}, {}, {})".format(func_name,
                                                    before_name,
                                                    after_name)


def _linter_function(function, before_all=None, after_all=None):
    """Return a _LinterFunction with before_all and after_all defaulted."""
    return _LinterFunction(function, before_all, after_all)

LINTER_FUNCTIONS = {
    "headerblock/filename": _linter_function(_filename_in_headerblock),
    "headerblock/desc_space": _linter_function(_space_in_headerblock),
    "headerblock/space_copyright": _linter_function(_space_before_copyright),
    "headerblock/copyright": _linter_function(_copyright_end_of_headerblock),
    "file/newline_last_char": _linter_function(_newline_end_of_file),
    "file/trailing_whitespace": _linter_function(_no_trailing_whitespace),
    "file/spelling_error": _linter_function(_no_spelling_errors,
                                            _construct_user_dictionary,
                                            _maybe_log_technical_terms)
}


def lint(relative_path_to_file,
         contents,
         linter_functions,
         **kwargs):
    r"""Actually lints some file contents.

    relative_path_to_file should contain the relative path to the file being
    linted from the root source directory. Contents should be a raw string
    with \n's.
    """
    lines = contents.splitlines(True)

    errors = list()
    for (code, info) in linter_functions.items():
        error = info.function(relative_path_to_file, lines, kwargs)
        if error:
            if isinstance(error, list):
                errors.extend([(code, e) for e in error])
            else:
                errors.append((code, error))

    errors = [e for e in errors if not _error_is_suppressed(e[1], e[0], lines)]
    return sorted(errors, key=lambda e: e[1].line)


def linter_functions_from_filters(whitelist=None,
                                  blacklist=None):
    """Yield tuples of _LinterFunction matching whitelist but not blacklist."""
    def _keyvalue_pair_if(dictionary, condition):
        """Return a key-value pair in dictionary if condition matched."""
        return {
            k: v for (k, v) in dictionary.items() if condition(k)
        }

    def _check_list(check_list, cond):
        """Return function testing against a list if the list exists."""
        def _check_against_list(key):
            """Return true if list exists and condition passes."""
            return cond(check_list, key) if check_list is not None else True

        return _check_against_list

    linter_functions = LINTER_FUNCTIONS
    linter_functions = _keyvalue_pair_if(linter_functions,
                                         _check_list(whitelist,
                                                     lambda l, k: k in l))
    linter_functions = _keyvalue_pair_if(linter_functions,
                                         _check_list(blacklist,
                                                     lambda l, k: k not in l))

    for code, linter_function in linter_functions.items():
        yield (code, linter_function)


class ShowAvailableChecksAction(argparse.Action):  # pylint:disable=R0903
    """If --checks is encountered, just show available checks and exit."""

    def __call__(self, parser, namespace, values, option_string=None):
        """Show available checks on --checks."""
        if option_string == "--checks":
            sys.stdout.write("Available options are:\n")
            for item in LINTER_FUNCTIONS.keys():
                sys.stdout.write(" * {0}\n".format(item))

            sys.exit(0)


def _parse_arguments(arguments=None):
    """Return a parser context result."""
    dir_hash = hashlib.sha1(os.getcwd().encode("utf-8")).hexdigest()

    parser = argparse.ArgumentParser(description="""Lint for Polysquare """
                                     """style guide""")
    parser.add_argument("--checks",
                        nargs=0,
                        action=ShowAvailableChecksAction,
                        help="""list available checks""")
    parser.add_argument("files",
                        nargs="*",
                        metavar=("FILE"),
                        help="""read FILE""",
                        type=str)
    parser.add_argument("--whitelist",
                        nargs="*",
                        help="""list of checks that should only be run""",
                        default=None)
    parser.add_argument("--blacklist",
                        nargs="*",
                        help="""list of checks that should never be run""",
                        default=None)
    parser.add_argument("--fix-what-you-can",
                        action="store_const",
                        help="""fix errors automatically""",
                        const=True)
    parser.add_argument("--spellcheck-cache",
                        help="""path to spell-checking cache file""",
                        default=None)
    parser.add_argument("--log-technical-terms-to",
                        help="""path to file to log technical terms to""",
                        default=None)
    parser.add_argument("--stamp-file-path",
                        help="""path to directory to store cached results""",
                        default=os.path.join(tempfile.gettempdir(),
                                             "jobstamps",
                                             "spellcheck_linter",
                                             dir_hash))
    parser.add_argument("--block-regexps",
                        help="""Regular expressions to exclude from """
                             """all checks.""",
                        nargs="*",
                        default=None)

    return parser.parse_args(arguments)


def _report_lint_error(error, file_path):
    """Report a linter error."""
    line = error[1].line
    code = error[0]
    description = error[1].description
    sys.stdout.write("{0}:{1} [{2}] {3}\n".format(file_path,
                                                  line,
                                                  code,
                                                  description))


def _apply_replacement(error, found_file, file_lines):
    """Apply a single replacement."""
    fixed_lines = file_lines
    fixed_lines[error[1].line - 1] = error[1].replacement
    concatenated_fixed_lines = "".join(fixed_lines)

    # Only fix one error at a time
    found_file.seek(0)
    found_file.write(concatenated_fixed_lines)
    found_file.truncate()


def _should_use_multiprocessing(num_jobs):
    """Return true if multiprocessing should be enabled."""
    return (not os.getenv("DISABLE_MULTIPROCESSING", None) and
            multiprocessing.cpu_count() < num_jobs and
            multiprocessing.cpu_count() > 2)


class ReprQueue(object):
    """Queue that has a stable __repr__."""

    def __init__(self, queue):
        """Initialize with underlying queue."""
        super(ReprQueue, self).__init__()
        self._queue = queue

    def __repr__(self):
        """Class name of underlying queue."""
        return self._queue.__class__.__name__

    def put(self, item, block=True, timeout=None):
        """Put item into underlying queue."""
        return self._queue.put(item, block, timeout)

    def get(self, block=True, timeout=None):
        """Get item from underlying queue."""
        return self._queue.get(block, timeout)

    def empty(self):
        """Return empty status from underlying queue."""
        return self._queue.empty()


def _obtain_queue(num_jobs):
    """Return queue type most appropriate for runtime model.

    If we are using multiprocessing, that should be
    multiprocessing.Manager().Queue. If we are just using a
    single process, then use a normal queue type.
    """
    if _should_use_multiprocessing(num_jobs):
        return ReprQueue(multiprocessing.Manager().Queue())

    return ReprQueue(Queue())


def tool_options_from_global(global_options, num_jobs):
    """From an argparse namespace, get a dict of options for the tools."""
    internal_opt = ["whitelist", "blacklist", "fix_what_you_can"]
    queue_object = _obtain_queue(num_jobs)
    translate = defaultdict(lambda: (lambda x: x),
                            log_technical_terms_to=lambda _: queue_object)

    tool_options = OrderedDict()

    for key in sorted(global_options.keys()):
        if key not in internal_opt and global_options[key] is not None:
            tool_options[key] = translate[key](global_options[key])

    return tool_options


class FileLinterFailure(namedtuple("LinterFailure", "absolute_path failure")):
    """A class representing a linter failure as occurring on a file."""

    def __lt__(self, other):
        """Return true if self should be sorted less than other."""
        if self.absolute_path == other.absolute_path:
            return self.failure[1].line < other.failure[1].line

        return self.absolute_path < other.absolute_path


def _run_lint_on_file(file_path,
                      linter_functions,
                      tool_options,
                      fix_what_you_can):
    """Run each function in linter_functions on filename.

    If fix_what_you_can is specified, then the first error that has a
    possible replacement will be automatically fixed on this file.
    """
    with open(file_path, "r+") as found_file:
        file_contents = found_file.read()
        file_lines = file_contents.splitlines(True)
        try:
            errors = lint(file_path[len(os.getcwd()) + 1:],
                          file_contents,
                          linter_functions,
                          **tool_options)
        except RuntimeError as err:
            msg = ("""RuntimeError in processing """
                   """{0} - {1}""".format(file_path, str(err)))
            raise RuntimeError(msg)

        if fix_what_you_can:
            for error_index, error in enumerate(errors):
                if error[1].replacement is not None:
                    _apply_replacement(error, found_file, file_lines)
                    errors[error_index] = (error[0],
                                           LinterFailure(error[1].description +
                                                         " ... FIXED",
                                                         error[1].line,
                                                         error[1].replacement))
                    break

        return [FileLinterFailure(file_path, e) for e in errors]


def _run_lint_on_file_exceptions(file_path,
                                 linter_functions,
                                 tool_options,
                                 fix_what_you_can):
    """Run except function in :linter_functions: on file_path.

    This function calls through to _run_lint_on_file, but prints a traceback
    for any exceptions encountered. It is intended to be used with parmap.
    """
    try:
        return _run_lint_on_file(file_path,
                                 linter_functions,
                                 tool_options,
                                 fix_what_you_can)
    except Exception as exception:
        sys.stderr.write("""Encountered exception """
                         """when processing {}\n""".format(file_path))
        traceback.print_exc()
        raise exception


def _run_lint_on_file_stamped_args(file_path,  # suppress(too-many-arguments)
                                   stamp_file_path,
                                   log_technical_terms_to,
                                   linter_functions,
                                   tool_options,
                                   fix_what_you_can):
    """Return tuple of args and kwargs that function would be called with."""
    dictionary_path = os.path.abspath("DICTIONARY")
    dependencies = [file_path]

    if os.path.exists(dictionary_path):
        dependencies.append(dictionary_path)

    kwargs = OrderedDict()
    kwargs["jobstamps_dependencies"] = dependencies
    kwargs["jobstamps_cache_output_directory"] = stamp_file_path

    if log_technical_terms_to:
        kwargs["jobstamps_output_files"] = [log_technical_terms_to]

    return ((file_path,
             linter_functions,
             tool_options,
             fix_what_you_can),
            kwargs)


def _run_lint_on_file_stamped(*args, **kwargs):
    """Run linter functions on file_path, stamping in stamp_file_path."""
    stamp_args, stamp_kwargs = _run_lint_on_file_stamped_args(*args, **kwargs)

    return jobstamp.run(_run_lint_on_file_exceptions,
                        *stamp_args,
                        **stamp_kwargs)


def _ordered(generator, *args, **kwargs):
    """Sort keys of unordered_dict and store in OrderedDict."""
    unordered_dict = {k: v for k, v in generator(*args, **kwargs)}
    keys = sorted(list(dict(unordered_dict).keys()))
    result = OrderedDict()
    for key in keys:
        result[key] = unordered_dict[key]

    return result


def _any_would_run(func, filenames, *args, **kwargs):
    """True if a linter function would be called on any of filenames."""
    if os.environ.get("_POLYSQUARE_GENERIC_FILE_LINTER_NO_STAMPING", None):
        return True

    for filename in filenames:
        stamp_args, stamp_kwargs = _run_lint_on_file_stamped_args(filename,
                                                                  *args,
                                                                  **kwargs)
        dependency = jobstamp.out_of_date(func,
                                          *stamp_args,
                                          **stamp_kwargs)

        if dependency:
            return True

    return False


def main(arguments=None):   # suppress(unused-function)
    """Entry point for the linter."""
    result = _parse_arguments(arguments)
    linter_funcs = _ordered(linter_functions_from_filters,
                            result.whitelist,
                            result.blacklist)
    global_options = vars(result)
    tool_options = tool_options_from_global(global_options, len(result.files))
    any_would_run = _any_would_run(_run_lint_on_file_exceptions,
                                   result.files,
                                   result.stamp_file_path,
                                   result.log_technical_terms_to,
                                   linter_funcs,
                                   tool_options,
                                   result.fix_what_you_can)

    if any_would_run:
        for linter_function in linter_funcs.values():
            if linter_function.before_all:
                linter_function.before_all(global_options, tool_options)

        use_multiprocessing = _should_use_multiprocessing(len(result.files))
    else:
        use_multiprocessing = False

    if use_multiprocessing:
        mapper = parmap.map
    else:
        # suppress(E731)
        mapper = lambda f, i, *a: [f(*((x, ) + a)) for x in i]

    errors = list(itertools.chain(*mapper(_run_lint_on_file_stamped,
                                          result.files,
                                          result.stamp_file_path,
                                          result.log_technical_terms_to,
                                          linter_funcs,
                                          tool_options,
                                          result.fix_what_you_can)))

    for error in sorted(errors):
        _report_lint_error(error.failure, os.path.relpath(error.absolute_path))

    if any_would_run:
        for linter_funcs in linter_funcs.values():
            if linter_funcs.after_all:
                linter_funcs.after_all(global_options, tool_options)

    return len(errors)



# /test/test_spelling.py
#
# Test cases for spelling-specific code in
#
# Disable no-self-use in tests as all test methods must be
# instance methods and we don't necessarily have to use a matcher
# with them.
# pylint:disable=no-self-use
#
# See /LICENCE.md for Copyright information
"""Test the linter to ensure that each lint use-case triggers warnings."""

import atexit

import inspect

import os

import shutil

import tempfile

import time

from nose_parameterized import parameterized

from polysquarelinter import spelling
from polysquarelinter.spelling import SpellcheckError

from testtools import ExpectedException, TestCase
from testtools.matchers import (Contains,
                                Equals as TTEqMatcher,
                                MatchesListwise,
                                MatchesSetwise,
                                Not)


# Pychecker complains about the Equals matcher failing to override comparator
# so do that here
class Equals(TTEqMatcher):  # suppress(R0903)
    """Matcher which tests equality."""

    def __init__(self, matchee):
        """Forward matchee to parent class."""
        super(Equals, self).__init__(matchee)

    def comparator(self, expected, other):
        """Check that expected == other."""
        return other == expected


class WordCacheTestCase(TestCase):
    """A test case base class which keeps a word cache from start to finish."""

    @classmethod
    def setUpClass(cls):  # suppress(N802)
        """Create a temporary directory to store word graph caches."""
        # This is the name of the directory that we want to
        # place our files in.
        word_cache_dir = os.path.join(os.getcwd(), "wordcache")
        cls.cache_dir = tempfile.mkdtemp(prefix=word_cache_dir)
        atexit.register(shutil.rmtree, cls.cache_dir)


class TestDictionary(WordCacheTestCase):
    """Test case for spelling.Dictionary object."""

    def test_expected_english_words_corrections(self):
        """Expected corrections returned for misspelled english word."""
        dictionary = spelling.Dictionary(spelling.valid_words_set(),
                                         "english_words",
                                         self.__class__.cache_dir)
        result = dictionary.corrections("splelling")
        self.assertFalse(result.valid)
        self.assertThat(result.suggestions,
                        MatchesSetwise(Equals("spelling"),
                                       Equals("spellings"),
                                       Equals("spieling"),
                                       Equals("spilling"),
                                       Equals("swelling")))

    def test_true_on_valid_word(self):
        """True and empty list returned for valid english word."""
        dictionary = spelling.Dictionary(spelling.valid_words_set(),
                                         "english_words",
                                         self.__class__.cache_dir)
        result = dictionary.corrections("spelling")
        self.assertTrue(result.valid)
        self.assertEqual(result.suggestions, list())


class TestDictionaryWithCustomWords(WordCacheTestCase):
    """Test case for spelling.Dictionary object."""

    def __init__(self, *args, **kwargs):
        """Initialize instance variables for this test case."""
        super(TestDictionaryWithCustomWords, self).__init__(*args, **kwargs)
        self.user_dictionary_path = None

    def setUp(self):  # suppress(N802)
        """Create a dictionary with custom words in it."""
        super(TestDictionaryWithCustomWords, self).setUp()
        self._current_path = os.getcwd()
        dictionary_dir = os.path.join(self._current_path, "user_dictionary")
        self._user_dictionary_dir = tempfile.mkdtemp(prefix=dictionary_dir)
        self.addCleanup(shutil.rmtree, self._user_dictionary_dir)
        self.user_dictionary_path = os.path.join(self._user_dictionary_dir,
                                                 "DICTIONARY")
        self.addCleanup(os.chdir, self._current_path)

    def _user_dictionary_set(self):
        """Get words in user dictionary as set."""
        return spelling.read_dictionary_file(self.user_dictionary_path)

    def test_spellcheck_without_cache(self):
        """Run spellcheck without cache.

        This is a simple check to ensure that a dictionary can still
        be constructed in-memory.
        """
        valid_words = spelling.valid_words_set(self.user_dictionary_path,
                                               self._user_dictionary_set())
        dictionary = spelling.Dictionary(valid_words,
                                         "english_and_custom_words")
        dictionary.corrections("mzcustomword")

    def test_custom_words_included_in_corrections(self):
        """Custom words included in corrections."""
        with open(self.user_dictionary_path, "w") as user_dictionary:
            user_dictionary.write("mycustomword")

        valid_words = spelling.valid_words_set(self.user_dictionary_path,
                                               self._user_dictionary_set())
        dictionary = spelling.Dictionary(valid_words,
                                         "english_and_custom_words",
                                         cache=self.__class__.cache_dir)
        result = dictionary.corrections("mzcustomword")
        self.assertFalse(result.valid)
        self.assertThat(result.suggestions,
                        MatchesSetwise(Equals("mycustomword")))

    def test_match_with_custom_word(self):
        """Match true on hitting custom word."""
        with open(self.user_dictionary_path, "w") as user_dictionary:
            user_dictionary.write("mycustomword\n")

        valid_words = spelling.valid_words_set(self.user_dictionary_path,
                                               self._user_dictionary_set())
        dictionary = spelling.Dictionary(valid_words,
                                         "english_and_custom_words",
                                         [self.user_dictionary_path],
                                         self.__class__.cache_dir)
        result = dictionary.corrections("mycustomword")
        self.assertTrue(result.valid)
        self.assertEqual(result.suggestions, list())

    def test_invalidate_custom_word_cache(self):
        """Regenerate internal user dictionary on file change."""
        with open(self.user_dictionary_path, "w") as user_dictionary:
            user_dictionary.write("mycustomword\n")

        valid_words = spelling.valid_words_set(self.user_dictionary_path,
                                               self._user_dictionary_set())
        dictionary = spelling.Dictionary(valid_words,
                                         "english_and_custom_words",
                                         [self.user_dictionary_path],
                                         self.__class__.cache_dir)
        dictionary.corrections("mycustomword")

        # Reassign to valid_words and dictionary, such that the caches
        # are re-generated on disk
        time.sleep(1)
        spelling.clear_caches()
        with open(self.user_dictionary_path, "w") as user_dictionary:
            user_dictionary.write("myothercustomword\n")

        valid_words = spelling.valid_words_set(self.user_dictionary_path,
                                               self._user_dictionary_set())
        dictionary = spelling.Dictionary(valid_words,
                                         "english_and_custom_words",
                                         [self.user_dictionary_path],
                                         self.__class__.cache_dir)
        result = dictionary.corrections("myothercustomword")
        self.assertTrue(result.valid)
        self.assertEqual(result.suggestions, list())


class TestSplitSpellcheckableFromShadowContents(TestCase):
    """Test case for the spellcheckable_and_shadow_contents function."""

    @staticmethod
    def shadow_contents_to_string(shadow):
        """Convert shadow contents to a single string."""
        # suppress(PYC70,PYC90)
        return "".join(["".join([c for c in l if c != 0]) for l in shadow])

    def test_spellcheckable_chunks_not_in_shadow_contents(self):
        """Spellcheckable chunk is not found in shadow contents."""
        contents = "# spellcheckable\n shadow".splitlines()
        chunks, shadow = spelling.spellcheckable_and_shadow_contents(contents)

        self.assertThat(self.__class__.shadow_contents_to_string(shadow),
                        Not(Contains(chunks[0].data[0])))

    def test_repeated_comment_markers(self):
        """Handle repeated comment markers."""
        contents = "# spell # checkable\n shadow".splitlines()
        chunks, _ = spelling.spellcheckable_and_shadow_contents(contents)

        self.assertEqual(chunks[0].data[0], " spell # checkable")

    def test_comment_out_markers_occurring_before_in_points(self):
        """Handle comment out points occurring before in points."""
        contents = "/* */\n*/ spell /* checkable */\n shadow".splitlines()
        chunks, _ = spelling.spellcheckable_and_shadow_contents(contents)

        self.assertEqual(chunks[1].data[0], " checkable ")

    def test_single_quoted_regions_not_found_in_shadow_contents(self):
        """Single quoted chunk is not found in shadow contents."""
        contents = "# spellcheckable\n 'quoted' shadow".splitlines()
        _, shadow = spelling.spellcheckable_and_shadow_contents(contents)

        self.assertThat(self.__class__.shadow_contents_to_string(shadow),
                        Not(Contains("quoted")))

    def test_double_quoted_regions_not_found_in_shadow_contents(self):
        """Double quoted chunk is not found in shadow contents."""
        contents = "# spellcheckable\n \"quoted\" shadow".splitlines()
        _, shadow = spelling.spellcheckable_and_shadow_contents(contents)

        self.assertThat(self.__class__.shadow_contents_to_string(shadow),
                        Not(Contains("quoted")))

    @parameterized.expand(list(" .=[](){}<>"))
    def test_split_technical_words_from_shadow_contents(self, character):
        """Split technical words from shadow contents."""
        contents = "#\n first_sym{0}second_sym".format(character).splitlines()
        _, shadow = spelling.spellcheckable_and_shadow_contents(contents)

        self.assertThat(spelling.technical_words_from_shadow_contents(shadow),
                        MatchesSetwise(Equals("first_sym"),
                                       Equals("second_sym")))

    def test_block_out_regions_of_spellcheckable_chunks(self):
        """Block out regions of spellcheckable chunks."""
        contents = "# blocked(region) other\n other_contents".splitlines()
        block_regexes = [
            r"\bblocked\([^\s]*\)"
        ]
        chunks, _ = spelling.spellcheckable_and_shadow_contents(contents,
                                                                block_regexes)

        self.assertThat(chunks[0].data[0],
                        Not(Contains("blocked(region)")))


class TestSpellcheckOnRegion(WordCacheTestCase):
    """Test cases for spellcheck_region function."""

    @classmethod
    def get_dictionaries(cls, contents):
        """Return a tuple of two dictionaries.

        The first dictionary is a Dictionary of valid words. The second
        dictionary is a Dictionary of technical words.
        """
        unique_name = inspect.currentframe().f_back.f_code.co_name
        _, shadow = spelling.spellcheckable_and_shadow_contents(contents)
        valid_words = spelling.Dictionary(spelling.valid_words_set(),
                                          "english_words",
                                          cls.cache_dir)
        tch_set = spelling.technical_words_from_shadow_contents(shadow)
        technical_words = spelling.Dictionary(tch_set,
                                              "technical_words_" + unique_name,
                                              cls.cache_dir)

        return (valid_words, technical_words)

    def test_exception_on_failure_to_detect_comment_system(self):
        """RuntimeError raised when comment system can't be detected."""
        contents = "no comment system"
        with ExpectedException(RuntimeError):
            valid, technical = self.__class__.get_dictionaries(contents)
            spelling.spellcheck_region(contents, valid, technical)

    def test_allow_matching_technical_words(self):
        """Spellcheck passes if a technical word is matched."""
        contents = "# technical_word\n technical_word".splitlines()
        valid, technical = self.__class__.get_dictionaries(contents)

        errors = [e for e in spelling.spellcheck_region(contents,
                                                        valid,
                                                        technical)]
        self.assertEqual(errors, list())

    def test_error_on_technical_word_not_found(self):
        """Spellcheck error if a technical word used isn't found."""
        contents = "# technical_word\n".splitlines()
        valid, technical = self.__class__.get_dictionaries(contents)

        errors = [e for e in spelling.spellcheck_region(contents,
                                                        valid,
                                                        technical)]
        self.assertEqual(errors[0].error_type,
                         spelling.SpellcheckError.TechnicalWord)

    def test_error_on_ordinary_mispelling(self):
        """Spellcheck error if ordinary word is misspelled."""
        contents = "# splelling\n".splitlines()
        valid, technical = self.__class__.get_dictionaries(contents)

        errors = [e for e in spelling.spellcheck_region(contents,
                                                        valid,
                                                        technical)]
        self.assertEqual(errors[0].error_type,
                         spelling.SpellcheckError.InvalidWord)

    def test_no_error_on_ordinary_word_valid(self):
        """No spellcheck error if ordinary word is valid."""
        contents = "# spelling\n".splitlines()
        valid, technical = self.__class__.get_dictionaries(contents)

        errors = [e for e in spelling.spellcheck_region(contents,
                                                        valid,
                                                        technical)]
        self.assertEqual(0, len(errors))

    def test_ignore_words_with_non_code_symbols(self):
        """Ignore words with non-code symbols in them."""
        contents = "# ~spelling\n".splitlines()
        valid, technical = self.__class__.get_dictionaries(contents)

        errors = [e for e in spelling.spellcheck_region(contents,
                                                        valid,
                                                        technical)]
        self.assertEqual(0, len(errors))

    def test_ignore_standalone_code_symbols(self):
        """Ignore standalone code-like symbols."""
        contents = "# : \n".splitlines()
        valid, technical = self.__class__.get_dictionaries(contents)

        errors = [e for e in spelling.spellcheck_region(contents,
                                                        valid,
                                                        technical)]
        self.assertEqual(0, len(errors))

    def test_always_allow_user_specified_word(self):
        """Allow user specified words even if they appear technical."""
        contents = "# a_technical\n".splitlines()
        valid, technical = self.__class__.get_dictionaries(contents)

        errors = [e for e in spelling.spellcheck_region(contents,
                                                        valid,
                                                        technical,
                                                        set(["a_technical"]))]
        self.assertEqual(0, len(errors))

    def test_error_on_ordinary_mispelling_no_technical_dictionary(self):
        """Error if ordinary word is misspelled, no technical dictionary."""
        contents = "# splelling\n".splitlines()
        valid, _ = self.__class__.get_dictionaries(contents)

        errors = [e for e in spelling.spellcheck_region(contents,
                                                        valid,
                                                        None)]
        self.assertEqual(errors[0].error_type,
                         spelling.SpellcheckError.InvalidWord)

    @parameterized.expand(["\"", "'", "(", "<", "{", "["])
    def test_handle_words_after_punctuation(self, punctuation):
        """Handle misspelled word after valid punctuation."""
        contents = "# ({0}missplled\n".format(punctuation).splitlines()
        valid, technical = self.__class__.get_dictionaries(contents)

        errors = [e for e in spelling.spellcheck_region(contents,
                                                        valid,
                                                        technical)]
        self.assertEqual(errors[0].error_type,
                         spelling.SpellcheckError.InvalidWord)

    @parameterized.expand(["\"", "'", ")", ">", "}", "]"])
    def test_handle_words_before_punctuation(self, punctuation):
        """Handle misspelled word before valid punctuation."""
        contents = "# missplled{0})\n".format(punctuation).splitlines()
        valid, technical = self.__class__.get_dictionaries(contents)

        errors = [e for e in spelling.spellcheck_region(contents,
                                                        valid,
                                                        technical)]
        self.assertEqual(errors[0].error_type,
                         spelling.SpellcheckError.InvalidWord)

    @parameterized.expand([":", ",", ":", ";"])
    def test_handle_words_before_punctuation_before_space(self, punctuation):
        """Handle misspelled word before valid punctuation (before space)."""
        contents = "# missplled{0} \n".format(punctuation).splitlines()
        valid, technical = self.__class__.get_dictionaries(contents)

        errors = [e for e in spelling.spellcheck_region(contents,
                                                        valid,
                                                        technical)]
        self.assertEqual(errors[0].error_type,
                         spelling.SpellcheckError.InvalidWord)

    def test_no_suggesting_technical_words(self):
        """Do not suggest technical words on ordinary word misspelling."""
        contents = "# splelling\n spl_lling".splitlines()
        valid, technical = self.__class__.get_dictionaries(contents)

        errors = [e for e in spelling.spellcheck_region(contents,
                                                        valid,
                                                        technical)]
        self.assertThat(errors[0].suggestions,
                        Not(Contains("spl_lling")))


class TestFilterNonspellcheckableTokens(TestCase):
    """Test cases for filtering out non-spellcheckable tokens."""

    @parameterized.expand(["https://domain.com/u/r/l.ext",
                           "path/to/file",
                           "/path/to/file",
                           "sample.email@address.com"])
    def test_filter_token(self, token):
        """Block out identifiers from text."""
        text = "valid {token} valid".format(token=token)
        blocked = spelling.filter_nonspellcheckable_tokens(text)
        self.assertEqual(blocked,
                         "valid {0} valid".format(" " * len(token)))


class TestSpellcheckErrors(TestCase):
    """Test spellcheck errors."""

    def test_sort_by_line_number(self):
        """Sort SpellcheckError by line number."""
        errors = [
            SpellcheckError("w", 2, 2, list(), SpellcheckError.InvalidWord),
            SpellcheckError("w", 1, 2, list(), SpellcheckError.TechnicalWord)
        ]

        self.assertThat(sorted(errors),
                        MatchesListwise([Equals(e) for e in [errors[1],
                                                             errors[0]]]))

    def test_sort_by_col_number(self):
        """Sort SpellcheckError by column number."""
        errors = [
            SpellcheckError("w", 1, 2, list(), SpellcheckError.InvalidWord),
            SpellcheckError("w", 1, 1, list(), SpellcheckError.TechnicalWord)
        ]

        self.assertThat(sorted(errors),
                        MatchesListwise([Equals(e) for e in [errors[1],
                                                             errors[0]]]))



# /test/test_warnings.py
#
# Test cases for each warning in polysquare-generic-file-linter
#
# Disable no-self-use in tests as all test methods must be
# instance methods and we don't necessarily have to use a matcher
# with them.
# pylint:disable=no-self-use
#
# See /LICENCE.md for Copyright information
"""Test the linter to ensure that each lint use-case triggers warnings."""

import atexit

import os

import shutil

import tempfile

from collections import defaultdict, namedtuple

from nose_parameterized import param, parameterized

from polysquarelinter import linter

from testtools import (ExpectedException, TestCase)


class LinterFailure(Exception):
    """Exception raised when the linter reports a message."""

    def __init__(self, message, repl):
        """Initialize exception with mesh and replacement."""
        super(LinterFailure, self).__init__()
        self.message = message
        self.replacement = repl

    def __str__(self):
        """Represent as string."""
        return str("{0}".format(self.message))

FormatStyle = namedtuple("FormatStyle", "start mid end")
_KNOWN_STYLES = [
    param(FormatStyle(start="#", mid="#", end="")),
    param(FormatStyle(start="rem", mid="rem", end="")),
    param(FormatStyle(start="//", mid="//", end="")),
    param(FormatStyle(start="/*", mid=" *", end=" */"))
]


def style_format(script, style):
    """Format script for style, replace keys in script with comments."""
    return script.format(s=style.start, m=style.mid, e=style.end)


def run_linter_throw(relative_path,
                     contents,
                     style,
                     **kwargs):
    """Run linter.lint and throws if it reports a message."""
    kwargs = defaultdict(lambda: None, **kwargs)
    functions = dict(linter.linter_functions_from_filters(kwargs["whitelist"],
                                                          kwargs["blacklist"]))
    tool_options = linter.tool_options_from_global(kwargs, 1)
    errors = linter.lint(relative_path,
                         style_format(contents, style),
                         functions,
                         **dict(tool_options))

    if len(errors):
        description = errors[0][1].description
        raise LinterFailure("{0}:{1} - {2} [{3}]".format(relative_path,
                                                         errors[0][1].line,
                                                         description,
                                                         errors[0][0]),
                            (errors[0][1].line, errors[0][1].replacement))

    return True


# Needed to silence pychecker warning - pychecker doesn't detect
# that self.assertRaises was overridden and thinks that the exception
# object returned is not the exception itself but unittest's old
# _ExceptionContext object # suppress(file/spelling_error)
def replacement(exception):
    """Get replacement stored in exception."""
    assert exception.__class__.__name__ == "LinterFailure"
    return exception.replacement


class DisableStampingTestCase(TestCase):
    """Base TestCase class that explicitly disables job caching."""

    def setUp(self):  # suppress(N802)
        """Disable job caching."""
        super(DisableStampingTestCase, self).setUp()
        os.environ["_POLYSQUARE_GENERIC_FILE_LINTER_NO_STAMPING"] = "1"


class TestFilenameHeaderWarnings(DisableStampingTestCase):
    """Test case for file names being at the top of a header."""

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_pass(self, style):
        """Check that headerblock/filename passes.

        Test passes where /path/to/file is in the header on the first line
        """
        result = run_linter_throw("path/to/file",
                                  "{s} /path/to/file\n{m}{e}\n",
                                  style,
                                  whitelist=["headerblock/filename"])
        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_ignore_shebangs(self, style):
        """Check that headerblock/filename passes, ignoring shebangs.

        Test passes where /path/to/file is in the header on the first line
        """
        result = run_linter_throw("path/to/file",
                                  "#!/usr/bin/env bash\n"
                                  "{s} /path/to/file\n{m}{e}\n",
                                  style,
                                  whitelist=["headerblock/filename"])
        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_ignore_at_echo(self, style):
        """Check that headerblock/filename passes, ignoring @ echo off.

        Test passes where /path/to/file is in the header on the first line
        """
        result = run_linter_throw("path/to/file",
                                  "@echo off\n"
                                  "{s} /path/to/file\n{m}{e}\n",
                                  style,
                                  whitelist=["headerblock/filename"])
        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_fail_malformed(self, style):
        """Check that headerblock/filename fails.

        Test fails where /path/to/file is not in the header on the first line
        """
        with ExpectedException(LinterFailure):
            run_linter_throw("path/to/file",
                             "{s} path/to/file_wrong\n{m}{e}\n",
                             style,
                             whitelist=["headerblock/filename"])

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_fail_nocomment(self, style):
        """Check that headerblock/filename fails.

        Test fails where /path/to/file is not in the header on the first line
        """
        with ExpectedException(RuntimeError):
            run_linter_throw("path/to/file",
                             "aabb\nbbcc",
                             style,
                             whitelist=["headerblock/filename"])

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_fail_short(self, style):
        """Check that headerblock/filename fails.

        Test fails where there are no lines
        """
        with ExpectedException(LinterFailure):
            run_linter_throw("path/to/file",
                             "",
                             style,
                             whitelist=["headerblock/filename"])

    @parameterized.expand(_KNOWN_STYLES)
    def test_suggest_filename(self, style):
        """Suggest the file name on headerblock/filename failure."""
        def get_replacement():
            """Get replacement for first line of headerblock."""
            run_linter_throw("path/to/file",
                             "{s}\n{m} Text{e}",
                             style,
                             whitelist=["headerblock/filename"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, style_format("{s} /path/to/file\n", style)))


class TestSpaceBetweenHeaderAndDescWarnings(DisableStampingTestCase):
    """Test case for a single blank comment between top and body of header."""

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_pass(self, style):
        """Check that headerblock/desc_space passes.

        Test passes where there is a single blank comment on the second line
        """
        result = run_linter_throw("path/to/file",
                                  "{s} /path/to/file\n{m}\n{m} Text{e}",
                                  style,
                                  whitelist=["headerblock/desc_space"])
        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_ignore_shebangs(self, style):
        """Check that headerblock/desc_space passes, ignoring shebangs.

        Test passes where there is a single blank comment on the second line
        """
        result = run_linter_throw("path/to/file",
                                  "#!/usr/bin/env bash\n"
                                  "{s} /path/to/file\n{m}\n{m} Text{e}",
                                  style,
                                  whitelist=["headerblock/desc_space"])
        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_ignore_at_echo(self, style):
        """Check that headerblock/desc_space passes, ignoring @ echo off.

        Test passes where there is a single blank comment on the second line
        """
        result = run_linter_throw("path/to/file",
                                  "@echo off\n"
                                  "{s} /path/to/file\n{m}\n{m} Text{e}",
                                  style,
                                  whitelist=["headerblock/desc_space"])
        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_fail_malformed(self, style):
        """Check that headerblock/desc_space fails.

        Test fail where there is not a single blank comment on the second line
        """
        with ExpectedException(LinterFailure):
            run_linter_throw("path/to/file",
                             "{s}\n{m} Text{e}",
                             style,
                             whitelist=["headerblock/desc_space"])

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_fail_short(self, style):
        """Check that headerblock/desc_space fails.

        Test fail where there are not even two lines
        """
        with ExpectedException(LinterFailure):
            run_linter_throw("path/to/file",
                             "{s}{e}\n",
                             style,
                             whitelist=["headerblock/desc_space"])

    @parameterized.expand(_KNOWN_STYLES)
    def test_suggest_insert_break(self, style):
        """Suggest a blank comment line on headerblock/desc_space failure."""
        def get_replacement():
            """Get replacement for lack of break."""
            run_linter_throw("path/to/file",
                             "{s}\n{m} Text{e}",
                             style,
                             whitelist=["headerblock/desc_space"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (2, style_format("{m}\n{m} Text{e}", style)))


class TestSpaceDescAndCopyrightWarnings(DisableStampingTestCase):
    """Test case for a single blank comment between bottom and body."""

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_pass(self, style):
        """Check that headerblock/space_copyright passes.

        Test passes where there is a single blank comment on the second
        last line
        """
        result = run_linter_throw("path/to/file",
                                  "{s} /path/to/file\n{m}\n{m} Text\n\n",
                                  style,
                                  whitelist=["headerblock/space_copyright"])
        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_fail(self, style):
        """Check that headerblock/desc_space_copyright fails.

        Test fails where there is not a single blank comment on the second
        last line
        """
        with ExpectedException(LinterFailure):
            run_linter_throw("path/to/file",
                             "{s} Text\n{m} Text\n {m} Text\n{e}\n",
                             style,
                             whitelist=["headerblock/space_copyright"])

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_fail_no_headerblock(self, style):
        """RuntimeError where file does not have headerblock."""
        with ExpectedException(RuntimeError):
            run_linter_throw("path/to/file",
                             "\n",
                             style,
                             whitelist=["headerblock/space_copyright"])

    @parameterized.expand(_KNOWN_STYLES)
    def test_suggest_insert_break(self, style):
        """Suggest a blank comment line for headerblock/space_copyright."""
        def get_replacement():
            """Get replacement for lack of break."""
            run_linter_throw("path/to/file",
                             "{s} Text\n{m} Text\n{m} Text{e}\n\n",
                             style,
                             whitelist=["headerblock/space_copyright"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (3, style_format("{m}\n{m} Text{e}\n", style)))


class TestCopyrightNotice(DisableStampingTestCase):
    """Test case for Copyright notice at end of header block."""

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_pass(self, style):
        """Check that headerblock/copyright passes.

        Test passes where "See /LICENCE.md for Copyright information" appears
        at the end of the header block
        """
        result = run_linter_throw("path/to/file",
                                  "{s} /path/to/file\n{m}\n{m} "
                                  "See /LICENCE.md for Copyright "
                                  "information{e}\n\n",
                                  style,
                                  whitelist=["headerblock/copyright"])
        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_fail(self, style):
        """Check that headerblock/copyright fails.

        Test fails where "See /LICENCE.md for Copyright information" does not
        appear at the end of the header block
        """
        with ExpectedException(LinterFailure):
            run_linter_throw("path/to/file",
                             "{s} /path/to/file\n"
                             "{m}\n{m} No Copyright Notice{e}\n\n",
                             style,
                             whitelist=["headerblock/copyright"])

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_fail_no_end(self, style):
        """headerblock/copyright fails where headerblock has no ending."""
        with ExpectedException(RuntimeError):
            run_linter_throw("path/to/file",
                             "{s} /path/to/file\n{m}\n{m}{e}",
                             style,
                             whitelist=["headerblock/copyright"])

    @parameterized.expand(_KNOWN_STYLES)
    def test_suggest_replacement(self, style):
        """Check a replacement is suggested if line has LICENCE or Copyright.

        The entire line should be replaced where it has LICENCE or Copyright
        as the user probably intended (but with a typographical error) to write
        the copyright notice
        """
        def get_replacement():
            """Get replacement for partial Copyright notice."""
            run_linter_throw("path/to/file",
                             "{s} /path/to/file\n"
                             "{m}\n{m} No Copyright Notice{e}\n\n",
                             style,
                             whitelist=["headerblock/copyright"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (3, style_format("{m} See /LICENCE.md for Copyright "
                                          "information{e}\n", style)))

    @parameterized.expand(_KNOWN_STYLES)
    def test_suggest_newline(self, style):
        """Check a new line is suggested where line has no notice.

        A new line should be suggested as no Copyright notice was even inserted
        in the first place
        """
        def get_replacement():
            """Get replacement for lack of Copyright notice."""
            run_linter_throw("path/to/file",
                             "{s} /path/to/file\n{m}\n{m} Other{e}\n\n",
                             style,
                             whitelist=["headerblock/copyright"])

        expected_repl = ("{m} Other\n"
                         "{m} See /LICENCE.md for Copyright information{e}\n")
        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (3, style_format(expected_repl, style)))


class TestNewlineAsLastChar(DisableStampingTestCase):
    r"""Test case for \n as last char of file."""

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_pass(self, style):
        r"""Check that file/newline_last_char passes.

        Test passes where "\n" is the last character in the file
        """
        result = run_linter_throw("path/to/file",
                                  "{s}\n{m}{e}\n",
                                  style,
                                  whitelist=["file/newline_last_char"])
        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_lint_fail(self, style):
        r"""Check that file/newline_last_char false.

        Test fails where "\n" is not the last character in the file
        """
        with ExpectedException(LinterFailure):
            run_linter_throw("path/to/file",
                             "{s}\n{m}{e}",
                             style,
                             whitelist=["file/newline_last_char"])

    @parameterized.expand(_KNOWN_STYLES)
    def test_suggest_newline(self, style):
        """Suggest a newline on the last line for file/newline_last_char."""
        def get_replacement():
            """Get replacement for lack of trailing newline."""
            run_linter_throw("path/to/file",
                             "{s}\n{m} Text{e}",
                             style,
                             whitelist=["file/newline_last_char"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (2, style_format("{m} Text{e}\n", style)))


class TestTrailingWhitespace(DisableStampingTestCase):
    """Test case for trailing whitespace."""

    def test_lint_pass(self):
        """File passes when there is no trailing whitespace."""
        contents = "contents\ncontents\n"
        result = run_linter_throw("path/to/file",
                                  contents,
                                  FormatStyle("#", "#", ""),
                                  whitelist=["file/trailing_whitespace"])
        self.assertTrue(result)

    @parameterized.expand(["contents     \ncontents\n",
                           "contents\n    \ncontents"])
    def test_lint_fail(self, contents):
        """File fails when there is trailing whitespace."""
        with ExpectedException(LinterFailure):
            run_linter_throw("path/to/file",
                             contents,
                             FormatStyle("#", "#", ""),
                             whitelist=["file/trailing_whitespace"])

    @parameterized.expand(["contents     \ncontents\n",
                           "    \ncontents"])
    def test_suggest_no_trailing_whitespace(self, contents):
        """File fails when there is trailing whitespace."""
        def get_replacement():
            """Get replacement for trailing newline."""
            run_linter_throw("path/to/file",
                             contents,
                             FormatStyle("#", "#", ""),
                             whitelist=["file/trailing_whitespace"])

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (1, contents.splitlines(True)[0].replace(" ", "")))


class TestSpellingErrors(DisableStampingTestCase):
    """Test case for spelling errors."""

    @classmethod
    def setUpClass(cls):  # suppress(N802)
        """Create a temporary directory to store word graph caches."""
        # This is the name of the directory that we want to
        # place our files in.
        word_cache_dir = os.path.join(os.getcwd(), "wordcache")
        cls.cache_dir = tempfile.mkdtemp(prefix=word_cache_dir)
        atexit.register(shutil.rmtree, cls.cache_dir)

    def _spellcheck_lint(self, text, style):
        """Wrap run_linter_throw and only run spellcheck.

        This function also provides the path to the word cache file which
        should speed up operation substantially.
        """
        cache_dir = self.__class__.cache_dir
        return run_linter_throw("path/to/file",
                                text,
                                style,
                                whitelist=["file/spelling_error"],
                                options={
                                    "spellcheck_cache": cache_dir
                                })

    @parameterized.expand(_KNOWN_STYLES)
    def test_spelling_mistake_inside_comment(self, style):
        """Find spelling errors inside comment."""
        with ExpectedException(LinterFailure):
            self._spellcheck_lint("{s}\n{m} splelling mistake{e}", style)

    @parameterized.expand(_KNOWN_STYLES)
    def test_no_spelling_error_comments_starting_inside_string(self, style):
        """No spelling errors where comment block starts inside string."""
        result = self._spellcheck_lint("{s}{e}\n\"{s}\" splelling",
                                       style)
        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_no_spelling_error_inside_double_quoted_string(self, style):
        """No spelling errors inside single double-quote string."""
        result = self._spellcheck_lint("{s}{e}\n\"splelling\"",
                                       style)
        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_no_spelling_error_inside_single_quoted_string(self, style):
        """No spelling errors inside single single-quote string."""
        result = self._spellcheck_lint("{s}{e}\n'splelling'",
                                       style)
        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_suggest_correction_inside_comment(self, style):
        """Correct spelling errors inside comment."""
        def get_replacement():
            self._spellcheck_lint("{s}\n{m} splelling mistake{e}", style)

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (2, style_format("{m} spelling mistake{e}",
                                          style)))

    @parameterized.expand(_KNOWN_STYLES)
    def test_spelling_mistake_in_double_quote_docstring(self, style):
        """Find spelling errors on double quoted string."""
        with ExpectedException(LinterFailure):
            self._spellcheck_lint("{s}{e}\ncall(\"\"\"splelling\"\"\")\n",
                                  style)

    @parameterized.expand(_KNOWN_STYLES)
    def test_suggest_correction_double_quote_docstring(self, style):
        """Correct spelling errors inside double quoted docstring."""
        def get_replacement():
            self._spellcheck_lint("{s}{e}\ncall(\"\"\"splelling\"\"\")\n",
                                  style)

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (2,
                          style_format("call(\"\"\"spelling\"\"\")\n",
                                       style)))

    @parameterized.expand(_KNOWN_STYLES)
    def test_spelling_mistake_in_multiline_docstring(self, style):
        """Find spelling errors in multi-line docstring."""
        with ExpectedException(LinterFailure):
            self._spellcheck_lint("{s}{e}\ncall(\"\"\"splelling\n"
                                  "mistake\"\"\")\n",
                                  style)

    @parameterized.expand(_KNOWN_STYLES)
    def test_suggest_correction_multiline_docstring(self, style):
        """Correct spelling errors in multi-line docstring."""
        def get_replacement():
            self._spellcheck_lint("{s}{e}\ncall(\"\"\"splelling\n"
                                  "mistake\"\"\")\n",
                                  style)

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (2,
                          style_format("call(\"\"\"spelling\n",
                                       style)))

    @parameterized.expand(_KNOWN_STYLES)
    def test_spelling_mistake_in_docstring(self, style):
        """Find spelling errors on docstring."""
        with ExpectedException(LinterFailure):
            self._spellcheck_lint("{s}{e}\n\"\"\"splelling mistake\"\"\")\n",
                                  style)

    @parameterized.expand(_KNOWN_STYLES)
    def test_spelling_mistake_for_technical_terms_not_in_context(self, style):
        """Technical terms not in same indent context are not valid words."""
        with ExpectedException(LinterFailure):
            content = "{s}{e}\n\"\"\"technicalterm\"\"\""
            self._spellcheck_lint(content, style)

    @parameterized.expand(_KNOWN_STYLES)
    def test_no_spelling_mistakes_for_technical_terms_in_context(self, style):
        """Technical terms used in rest of file are valid words."""
        content = "{s}{e}\n\"\"\"technicalterm\"\"\")\n\n technicalterm\n"
        result = self._spellcheck_lint(content, style)

        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_spelling_mistake_for_referenced_symbol_not_in_context(self,
                                                                   style):
        """Technical terms not referenced in context are invalid."""
        with ExpectedException(LinterFailure):
            content = "{s}{e}\n\"\"\"@technical_term\"\"\""
            self._spellcheck_lint(content, style)

    @parameterized.expand(_KNOWN_STYLES)
    def test_no_spelling_mistakes_for_referenced_symbol_in_context(self,
                                                                   style):
        """Technical terms used in rest of file are valid words."""
        content = "{s}{e}\n\"\"\"@technical_term\"\"\")\n\n technical_term\n"
        result = self._spellcheck_lint(content, style)

        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_no_spelling_mistakes_after_docstring(self, style):
        """No spelling errors after docstring."""
        content = "{s}{e}\n\"\"\"spelling mistake\"\"\")\n\n msspltword\n"
        result = self._spellcheck_lint(content, style)

        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_apostrophe_in_comments(self, style):
        """Handle apostrophes in comments correctly."""
        content = "{s}There's{e}\n msspltword\n{s} We're {e}"
        result = self._spellcheck_lint(content, style)

        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_comments_in_quotes(self, style):
        """Handle comments in quotes correctly."""
        content = "{s}{e}\n\"{s} There{e}\"\n msspltword\n"
        result = self._spellcheck_lint(content, style)

        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_comments_in_multiline_quotes(self, style):
        """Handle comments in multi-line quotes correctly."""
        content = "{s}{e}\n\"{s} There\n{e}\"\n msspltword\n"
        result = self._spellcheck_lint(content, style)

        self.assertTrue(result)

    @parameterized.expand(_KNOWN_STYLES)
    def test_spelling_mistake_in_single_quote_string(self, style):
        """Find spelling errors on single quoted docstring."""
        with ExpectedException(LinterFailure):
            self._spellcheck_lint("{s}{e}\ncall('''splelling mistake''')\n",
                                  style)

    @parameterized.expand(_KNOWN_STYLES)
    def test_suggest_correction_single_quote_string(self, style):
        """Correct spelling errors inside single quoted docstring."""
        def get_replacement():
            self._spellcheck_lint("{s}{e}\ncall('''splelling mistake''')\n",
                                  style)

        exception = self.assertRaises(LinterFailure, get_replacement)
        self.assertEqual(replacement(exception),
                         (2, style_format("call('''spelling mistake''')\n",
                                          style)))



# /polysquarelinter/lint_spelling_only.py
#
# Entry point for a special linter that only handles spelling errors.
#
# See /LICENCE.md for Copyright information
"""Entry point for a special linter that only handles spelling errors."""

import argparse

import hashlib

import os

import sys

import tempfile

from jobstamps import jobstamp

from polysquarelinter import spelling
from polysquarelinter.spelling import Dictionary, spellcheck_region


def spellcheck(contents, technical_terms=None, spellcheck_cache=None):
    """Run spellcheck on the contents of a file.

    :technical_terms: is a path to a file containing a list of "technical"
    terms. These may be symbols as collected from files by using
    the generic linter or other such symbols. If a symbol-like term is
    used within contents and it does not appear in :technical_terms: then
    an error will result.

    :spellcheck_cache: is a path to a directory where graph files generated
    by the spellchecking engine should be stored. It is used for caching
    purposes between invocations, since generating the spellchecking
    graph is an expensive operation which can take a few seconds to complete.
    """
    contents = spelling.filter_nonspellcheckable_tokens(contents)
    lines = contents.splitlines(True)
    user_dictionary = os.path.join(os.getcwd(), "DICTIONARY")
    user_words = spelling.read_dictionary_file(user_dictionary)

    valid_words = Dictionary(spelling.valid_words_set(user_dictionary,
                                                      user_words),
                             "valid_words",
                             dictionary_sources=[user_dictionary],
                             cache=spellcheck_cache)

    # By default, include the user dictionary in the technical terms available.
    technical_terms_set = user_words

    if technical_terms:
        with open(technical_terms) as tech_tf:
            technical_terms_set |= set(tech_tf.read().splitlines())

    technical_words = Dictionary(technical_terms_set,
                                 "technical_words",
                                 dictionary_sources=[technical_terms,
                                                     user_dictionary],
                                 cache=spellcheck_cache)

    return sorted([e for e in spellcheck_region(lines,
                                                valid_words,
                                                technical_words,
                                                user_words)])


def _parse_arguments(arguments=None):
    """Return a parser context result."""
    dir_hash = hashlib.sha1(os.getcwd().encode("utf-8")).hexdigest()

    parser = argparse.ArgumentParser(description="""Find spelling errors""")
    parser.add_argument("files",
                        nargs="*",
                        metavar=("FILE"),
                        help="""read FILE""",
                        type=str)
    parser.add_argument("--spellcheck-cache",
                        help="""path to spell-checking cache file""",
                        default=None)
    parser.add_argument("--technical-terms",
                        help="""path to file to source technical terms from""",
                        default=None)
    parser.add_argument("--stamp-file-path",
                        help="""path to directory to store cached results""",
                        default=os.path.join(tempfile.gettempdir(),
                                             "jobstamps",
                                             "spellcheck_linter",
                                             dir_hash))

    return parser.parse_args(arguments)


_SPELLCHECK_MESSAGES = {
    spelling.SpellcheckError.InvalidWord: """Misspelled word {0}""",
    spelling.SpellcheckError.TechnicalWord: """Technical term {0} not used """
                                            """in any scanned code file"""
}


def _report_spelling_error(error, file_path):
    """Report a spelling error."""
    line = error.line_offset + 1
    code = "file/spelling_error"
    description = _SPELLCHECK_MESSAGES[error.error_type].format(error.word)
    if error.suggestions is not None:
        description = (description +
                       ", perhaps you meant: " +
                       ", ".join(error.suggestions))

    sys.stdout.write("{0}:{1} [{2}] {3}\n".format(file_path,
                                                  line,
                                                  code,
                                                  description))


def main(arguments=None):  # suppress(unused-function)
    """Entry point for the spellcheck linter."""
    dictionary_path = os.path.abspath("DICTIONARY")
    result = _parse_arguments(arguments)

    num_errors = 0
    for found_filename in result.files:
        file_path = os.path.abspath(found_filename)
        with open(file_path, "r+") as found_file:
            jobstamps_dependencies = [file_path]

            if os.path.exists(dictionary_path):
                jobstamps_dependencies.append(dictionary_path)

            if (result.technical_terms and
                    os.path.exists(result.technical_terms)):
                jobstamps_dependencies.append(result.technical_terms)

            kwargs = {
                "jobstamps_dependencies": jobstamps_dependencies,
                "jobstamps_cache_output_directory": result.stamp_file_path,
            }

            errors = jobstamp.run(spellcheck,
                                  found_file.read(),
                                  result.technical_terms,
                                  result.spellcheck_cache,
                                  **kwargs)

            for error in errors:
                _report_spelling_error(error, file_path)

            num_errors += len(errors)

    return num_errors



# /test/test_linter_acceptance.py
#
# Acceptance tests for polysquarelinter.linter, actually runs
# main() and checks results of various functions.
#
# Disable no-self-use in tests as all test methods must be
# instance methods and we don't necessarily have to use a matcher
# with them.
# pylint:disable=no-self-use
#
# See /LICENCE.md for Copyright information
"""Test the linter to ensure that each lint use-case triggers warnings."""

import doctest

import errno

import os

import shutil

import sys

import tempfile

from contextlib import contextmanager

from iocapture import capture

from nose_parameterized import parameterized

from polysquarelinter import linter

from testtools import TestCase
from testtools.matchers import (DocTestMatches,
                                Equals as TTEqMatcher,
                                MatchesSetwise)


# Pychecker complains about the Equals matcher failing to override comparator
# so do that here
class Equals(TTEqMatcher):  # suppress(R0903)
    """Matcher which tests equality."""

    def __init__(self, matchee):
        """Forward matchee to parent class."""
        super(Equals, self).__init__(matchee)

    def comparator(self, expected, other):
        """Check that expected == other."""
        return other == expected


class LinterFailure(Exception):
    """Exception raised when the linter reports a message."""

    def __init__(self, message, repl):
        """Initialize exception with mesh and replacement."""
        super(LinterFailure, self).__init__()
        self.message = message
        self.replacement = repl

    def __str__(self):
        """Represent as string."""
        return str("{0}".format(self.message))


def run_with_kwargs_as_switches(func, *args, **kwargs):
    """Run :func: with :kwargs: converted to switches."""
    arguments = list(args)

    def _convert_kv_to_switches(key, value):
        """Convert a key-value pair to command-line switches."""
        append_args = ["--{0}".format(key).replace("_", "-")]

        type_dispatch = {
            bool: [],
            list: value,
            str: [value]
        }

        # We assume that the types in type_dispatch are the only types
        # we'll encounter, all others will throw an exception.
        append_args += type_dispatch[type(value)]
        return append_args

    for key, value in kwargs.items():
        arguments += _convert_kv_to_switches(key, value)

    return func(arguments)


def run_linter_main(filename, **kwargs):
    """Run linter.main() (as an integration test)."""
    arguments = [filename]
    return run_with_kwargs_as_switches(linter.main, *arguments, **kwargs)


class TestLinterAcceptance(TestCase):
    """Acceptance tests for linter.main()."""

    def __init__(self, *args, **kwargs):  # pylint:disable=super-on-old-class
        """Initialize class variables."""
        super(TestLinterAcceptance, self).__init__(*args, **kwargs)
        self._temporary_file = None

    def setUp(self):  # suppress(N802)
        """Create a temporary file."""
        from six import StringIO

        super(TestLinterAcceptance, self).setUp()
        self._temporary_file = tempfile.mkstemp()
        self.patch(sys, "stdout", StringIO())

        os.environ["JOBSTAMPS_DISABLED"] = "1"

    def tearDown(self):  # suppress(N802)
        """Remove temporary file.

        Note that we need to ensure that the file is closed
        first, so if it hasn't been opened yet, we won't get
        EBADF. Otherwise we'll get EBADF and we can safely
        ignore it.
        """
        try:
            os.close(self._temporary_file[0])
        except OSError as error:
            if error.errno != errno.EBADF:   # suppress(PYC90)
                raise error

        os.remove(self._temporary_file[1])
        super(TestLinterAcceptance, self).tearDown()

    def test_parallelization_path(self):
        """Generate expected number of errors when running in parallel."""
        contents = ("#\n"
                    "#\n"
                    "# Description\n"
                    "#\n"
                    "# See LICENCE.md for Copyright information\n"
                    "\n")

        temporary_dir = tempfile.mkdtemp(prefix=os.path.join(os.getcwd(),
                                                             "technical"))
        self.addCleanup(lambda: shutil.rmtree(temporary_dir))
        files_to_lint = []

        for i in range(0, 20):
            with open(os.path.join(temporary_dir,
                                   "file{0}".format(i)), "w") as lint_file:
                lint_file.write(contents)
                files_to_lint.append(os.path.realpath(lint_file.name))

        result = run_with_kwargs_as_switches(linter.main,
                                             *files_to_lint,
                                             whitelist="headerblock/copyright")

        self.assertEqual(result, 20)

    def test_inline_suppressions_above(self):
        """Check inline suppressions work above the error-generating line."""
        contents = ("#\n"
                    "#\n"
                    "# Description\n"
                    "#\n"
                    "# suppress(headerblock/copyright)\n"
                    "# See LICENCE.md for Copyright information\n"
                    "\n")

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        result = run_linter_main(self._temporary_file[1],
                                 whitelist=["headerblock/copyright"])

        self.assertEqual(result, 0)

    def test_inline_suppressions_beside(self):
        """Check inline suppressions work beside the error-generating line."""
        contents = ("#\n"
                    "#\n"
                    "# Description\n"
                    "#\n"
                    "# See LICENCE.md for Copyright information"
                    "# suppress(headerblock/copyright)\n"  # on the same line
                    "\n")

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        result = run_linter_main(self._temporary_file[1],
                                 whitelist=["headerblock/copyright"])

        self.assertEqual(result, 0)

    def test_blacklist(self):
        """Check that blacklisting a test causes it not to run."""
        contents = ("#\n"
                    "#\n"
                    "# Description\n"
                    "#\n"
                    "# See /LICENCE.md for Copyright information\n"
                    "\n")

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        result = run_linter_main(self._temporary_file[1],
                                 blacklist=["headerblock/filename",
                                            "file/spelling_error"])

        self.assertEqual(result, 0)

    def test_whitelist_pass(self):
        """Check that white-listing a test causes only it to run."""
        contents = ("#\n")

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        result = run_linter_main(self._temporary_file[1],
                                 whitelist=["file/newline_last_char"])

        self.assertEqual(result, 0)

    def test_whitelist_fail(self):
        """Check that white-listing a test causes only it to run."""
        contents = ("#")

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        result = run_linter_main(self._temporary_file[1],
                                 whitelist=["file/newline_last_char"])

        self.assertEqual(result, 1)

    def test_fix_what_you_can(self):
        """Check that --fix-what-you-can modifies file correctly."""
        contents = ("#")

        with os.fdopen(self._temporary_file[0], "a+") as process_file:
            process_file.write(contents)

        run_linter_main(self._temporary_file[1],
                        whitelist=["file/newline_last_char"],
                        fix_what_you_can=True)

        with open(self._temporary_file[1], "r") as processed_file:
            self.assertEqual("#\n", processed_file.read())

    @parameterized.expand([c for c in linter.LINTER_FUNCTIONS.keys()])
    def test_show_checks(self, check):
        """Check that --checks shows a specified check."""
        if check != list(linter.LINTER_FUNCTIONS.keys())[-1]:
            final_ellipsis = " ..."
        else:
            final_ellipsis = ""

        doctest_contents = ("... * {0}{1}").format(check, final_ellipsis)

        with capture() as captured:
            self.patch(sys, "exit", lambda _: None)
            linter.main(["--checks"])

            self.assertThat(captured.stdout,  # suppress(PYC70)
                            DocTestMatches(doctest_contents,
                                           doctest.ELLIPSIS |
                                           doctest.NORMALIZE_WHITESPACE |
                                           doctest.REPORT_NDIFF))

    def test_log_technical_words_over_two_files(self):
        """Check that --log-technical-words combines found technical words."""
        @contextmanager
        def in_dir(directory):
            """Perform actions in the context of directory."""
            last_directory = os.getcwd()
            os.chdir(directory)
            try:
                yield
            finally:
                os.chdir(last_directory)

        temporary_dir = tempfile.mkdtemp(prefix=os.path.join(os.getcwd(),
                                                             "technical"))
        self.addCleanup(lambda: shutil.rmtree(temporary_dir))

        with in_dir(temporary_dir):
            first_file_path = os.path.join(temporary_dir, "first_file.txt")
            second_file_path = os.path.join(temporary_dir, "second_file.txt")
            tech_terms_path = os.path.join(temporary_dir,
                                           "technical_terms.txt")

            with open(first_file_path, "w") as f:
                f.write("#\n technical_term_one shared_technical_term\n")

            with open(second_file_path, "w") as f:
                f.write("#\n technical_term_two shared_technical_term\n")

            run_with_kwargs_as_switches(linter.main,
                                        first_file_path,
                                        second_file_path,
                                        whitelist=["file/spelling_error"],
                                        log_technical_terms_to=tech_terms_path)

            with open(tech_terms_path, "r") as f:
                logged_technical_terms = f.read().splitlines()

            self.assertThat(logged_technical_terms,
                            MatchesSetwise(Equals("technical_term_one"),
                                           Equals("technical_term_two"),
                                           Equals("shared_technical_term")))



# /setup.py
#
# Installation and setup script for polysquare-generic-file-linter
#
# See /LICENCE.md for Copyright information
"""Installation and setup script for polysquare-generic-file-linter."""

from setuptools import find_packages
from setuptools import setup

setup(name="polysquare-generic-file-linter",
      version="0.1.7",
      description="Polysquare Style Guide Linter",
      long_description_markdown_filename="README.md",
      author="Sam Spilsbury",
      author_email="smspillaz@gmail.com",
      url="http://github.com/polysquare/polysquare-generic-file-linter",
      classifiers=["Development Status :: 3 - Alpha",
                   "Programming Language :: Python :: 2",
                   "Programming Language :: Python :: 2.7",
                   "Programming Language :: Python :: 3",
                   "Programming Language :: Python :: 3.1",
                   "Programming Language :: Python :: 3.2",
                   "Programming Language :: Python :: 3.3",
                   "Programming Language :: Python :: 3.4",
                   "Intended Audience :: Developers",
                   "Topic :: Software Development :: Build Tools",
                   "License :: OSI Approved :: MIT License"],
      license="MIT",
      keywords="development linters",
      packages=find_packages(exclude=["test"]),
      package_data={
          "polysquarelinter": [
              "polysquarelinter/en_US.txt"
          ]
      },
      install_requires=[
          "jobstamps>=0.0.16",
          "parmap",
          "Whoosh<=2.6.0"
      ],
      extras_require={
          "polysquarelint": ["polysquare-setuptools-lint>=0.0.22"],
          "green": [
              "iocapture",
              "nose",
              "nose-parameterized>=0.5.0",
              "setuptools-green>=0.0.11",
              "six",
              "testtools"
          ],
          "upload": ["setuptools-markdown"]
      },
      entry_points={
          "console_scripts": [
              "polysquare-generic-file-linter=polysquarelinter.linter:main",
              "spellcheck-linter=polysquarelinter.lint_spelling_only:main"
          ]
      },
      test_suite="nose.collector",
      zip_safe=True,
      include_package_data=True)



# /polysquarelinter/__init__.py
#
# Entry point for linter.
#
# See /LICENCE.md for Copyright information
"""Load the polysquarelinter package."""

import os

# Hack to work around a bug in pkg_resources.
#
# We need to override the __file__ attribute of polysquarelinter.__file__
# with its absolute path, as opposed to its relative path. This fixes an
# issue where running tests in the source directory and then changing
# directory mid-test will cause pkg_resources to fail, since it relies on
# __file__ to be a valid path (which ceases to be valid as soon as we
# exit the source directory).
if not os.path.isabs(__file__):
    __file__ = os.path.abspath(__file__)
